import logging
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    TYPE_CHECKING,
)

import tool_shed.util.repository_util
from galaxy.util import (
    asbool,
    listify,
)
from tool_shed.util import (
    common_util,
    container_util,
    metadata_util,
    shed_util_common as suc,
)
from tool_shed.webapp.model.db import get_repository_by_name_and_owner

if TYPE_CHECKING:
    from tool_shed.structured_app import ToolShedApp

log = logging.getLogger(__name__)


class RelationBuilder:
    def __init__(self, app: "ToolShedApp", repository, repository_metadata, tool_shed_url, trans=None):
        self.all_repository_dependencies: Dict[str, Any] = {}
        self.app = app
        self.circular_repository_dependencies: List[Tuple] = []
        self.repository = repository
        self.repository_metadata = repository_metadata
        self.handled_key_rd_dicts: List[Dict[str, List[str]]] = []
        self.key_rd_dicts_to_be_processed: List[Dict[str, List[str]]] = []
        self.tool_shed_url = tool_shed_url
        self.trans = trans

    def can_add_to_key_rd_dicts(self, key_rd_dict, key_rd_dicts: List[Dict[str, List[str]]]):
        """Handle the case where an update to the changeset revision was done."""
        k = next(iter(key_rd_dict))
        rd = key_rd_dict[k]
        partial_rd = rd[0:3]
        for kr_dict in key_rd_dicts:
            key = next(iter(kr_dict))
            if key == k:
                repository_dependency = kr_dict[key]
                if repository_dependency[0:3] == partial_rd:
                    return False
        return True

    def filter_only_if_compiling_contained_td(self, key_rd_dict: Dict[str, Any]):
        """
        Return a copy of the received key_rd_dict with repository dependencies that are needed
        only_if_compiling_contained_td filtered out of the list of repository dependencies for
        each rd_key.
        """
        filtered_key_rd_dict: Dict[str, Any] = {}
        for rd_key, required_rd_tup in key_rd_dict.items():
            (
                tool_shed,
                name,
                owner,
                changeset_revision,
                prior_installation_required,
                only_if_compiling_contained_td,
            ) = common_util.parse_repository_dependency_tuple(required_rd_tup)
            if not asbool(only_if_compiling_contained_td):
                filtered_key_rd_dict[rd_key] = required_rd_tup
        return filtered_key_rd_dict

    def get_prior_installation_required_and_only_if_compiling_contained_td(self):
        """
        This method is called from the tool shed and never Galaxy.  If self.all_repository_dependencies
        contains a repository dependency tuple that is associated with self.repository, return the
        value of the tuple's prior_installation_required component.
        """
        cleaned_toolshed_base_url = common_util.remove_protocol_from_tool_shed_url(self.tool_shed_url)
        if self.all_repository_dependencies:
            for rd_key, rd_tups in self.all_repository_dependencies.items():
                if rd_key in ["root_key", "description"]:
                    continue
                for rd_tup in rd_tups:
                    (
                        rd_toolshed,
                        rd_name,
                        rd_owner,
                        rd_changeset_revision,
                        rd_prior_installation_required,
                        rd_only_if_compiling_contained_td,
                    ) = common_util.parse_repository_dependency_tuple(rd_tup)
                    cleaned_rd_toolshed = common_util.remove_protocol_from_tool_shed_url(rd_toolshed)
                    if (
                        cleaned_rd_toolshed == cleaned_toolshed_base_url
                        and rd_name == self.repository.name
                        and rd_owner == self.repository.user.username
                        and rd_changeset_revision == self.repository_metadata.changeset_revision
                    ):
                        return rd_prior_installation_required, rd_only_if_compiling_contained_td
        elif self.repository_metadata:
            # Get the list of changeset revisions from the tool shed to which self.repository may be updated.
            metadata = self.repository_metadata.metadata
            current_changeset_revision = str(self.repository_metadata.changeset_revision)
            # Get the changeset revision to which the current value of required_repository_changeset_revision
            # should be updated if it's not current.
            text = metadata_util.get_updated_changeset_revisions(
                self.app,
                name=str(self.repository.name),
                owner=str(self.repository.user.username),
                changeset_revision=current_changeset_revision,
            )
            if text:
                valid_changeset_revisions = listify(text)
                if current_changeset_revision not in valid_changeset_revisions:
                    valid_changeset_revisions.append(current_changeset_revision)
            else:
                valid_changeset_revisions = [current_changeset_revision]
            repository_dependencies_dict = metadata["repository_dependencies"]
            rd_tups = repository_dependencies_dict.get("repository_dependencies", [])
            for rd_tup in rd_tups:
                (
                    rd_toolshed,
                    rd_name,
                    rd_owner,
                    rd_changeset_revision,
                    rd_prior_installation_required,
                    rd_only_if_compiling_contained_td,
                ) = common_util.parse_repository_dependency_tuple(rd_tup)
                cleaned_rd_toolshed = common_util.remove_protocol_from_tool_shed_url(rd_toolshed)
                if (
                    cleaned_rd_toolshed == cleaned_toolshed_base_url
                    and rd_name == self.repository.name
                    and rd_owner == self.repository.user.username
                    and rd_changeset_revision in valid_changeset_revisions
                ):
                    return rd_prior_installation_required, rd_only_if_compiling_contained_td
        # Default both prior_installation_required and only_if_compiling_contained_td to False.
        return "False", "False"

    def get_key_for_repository_changeset_revision(self):
        # The received toolshed_base_url must include the port, but doesn't have to include the protocol.
        (
            prior_installation_required,
            only_if_compiling_contained_td,
        ) = self.get_prior_installation_required_and_only_if_compiling_contained_td()
        # Create a key with the value of prior_installation_required defaulted to False.
        key = container_util.generate_repository_dependencies_key_for_repository(
            self.tool_shed_url,
            self.repository.name,
            self.repository.user.username,
            self.repository_metadata.changeset_revision,
            prior_installation_required,
            only_if_compiling_contained_td,
        )
        return key

    def get_repository_dependencies_for_changeset_revision(self):
        """
        Return a dictionary of all repositories upon which the contents of self.repository_metadata
        record depend.  The dictionary keys are name-spaced values consisting of:
        self.tool_shed_url/repository_name/repository_owner/changeset_revision
        and the values are lists of repository_dependency tuples consisting of:
        ( self.tool_shed_url, repository_name, repository_owner, changeset_revision ).
        This method ensures that all required repositories to the nth degree are returned.
        """
        # Assume the current repository does not have repository dependencies defined for it.
        current_repository_key: Optional[str] = None
        if metadata := self.repository_metadata.metadata:
            # The value of self.tool_shed_url must include the port, but doesn't have to include
            # the protocol.
            if "repository_dependencies" in metadata:
                current_repository_key = self.get_key_for_repository_changeset_revision()
                repository_dependencies_dict = metadata["repository_dependencies"]
                if not self.all_repository_dependencies:
                    self.initialize_all_repository_dependencies(current_repository_key, repository_dependencies_dict)
                # Handle the repository dependencies defined in the current repository, if any, and populate
                # the various repository dependency objects for this round of processing.
                current_repository_key_rd_dicts = self.populate_repository_dependency_objects_for_processing(
                    current_repository_key, repository_dependencies_dict
                )
        if current_repository_key:
            if current_repository_key_rd_dicts:
                # There should be only a single current_repository_key_rd_dict in this list.
                current_repository_key_rd_dict = current_repository_key_rd_dicts[0]
                # Handle circular repository dependencies.
                if not self.in_circular_repository_dependencies(current_repository_key_rd_dict):
                    if current_repository_key in self.all_repository_dependencies:
                        self.handle_current_repository_dependency(current_repository_key)
                elif self.key_rd_dicts_to_be_processed:
                    self.handle_next_repository_dependency()
            elif self.key_rd_dicts_to_be_processed:
                self.handle_next_repository_dependency()
        elif self.key_rd_dicts_to_be_processed:
            self.handle_next_repository_dependency()
        self.all_repository_dependencies = self.prune_invalid_repository_dependencies(self.all_repository_dependencies)
        return self.all_repository_dependencies

    def get_repository_dependency_as_key(self, repository_dependency):
        (
            tool_shed,
            name,
            owner,
            changeset_revision,
            prior_installation_required,
            only_if_compiling_contained_td,
        ) = common_util.parse_repository_dependency_tuple(repository_dependency)
        return container_util.generate_repository_dependencies_key_for_repository(
            tool_shed, name, owner, changeset_revision, prior_installation_required, only_if_compiling_contained_td
        )

    def get_updated_changeset_revisions_for_repository_dependencies(self, key_rd_dicts: List[Dict[str, Any]]):
        updated_key_rd_dicts: List[Dict[str, Any]] = []
        for key_rd_dict in key_rd_dicts:
            key = next(iter(key_rd_dict))
            repository_dependency = key_rd_dict[key]
            (
                rd_toolshed,
                rd_name,
                rd_owner,
                rd_changeset_revision,
                rd_prior_installation_required,
                rd_only_if_compiling_contained_td,
            ) = common_util.parse_repository_dependency_tuple(repository_dependency)
            tool_shed_is_this_tool_shed = suc.tool_shed_is_this_tool_shed(rd_toolshed, trans=self.trans)
            if tool_shed_is_this_tool_shed:
                repository = get_repository_by_name_and_owner(self.app.model.context, rd_name, rd_owner)
                if repository:
                    repository_id = self.app.security.encode_id(repository.id)
                    repository_metadata = metadata_util.get_repository_metadata_by_repository_id_changeset_revision(
                        self.app, repository_id, rd_changeset_revision
                    )
                    if repository_metadata:
                        # The repository changeset_revision is installable, so no updates are available.
                        new_key_rd_dict = {key: repository_dependency}
                        updated_key_rd_dicts.append(key_rd_dict)
                    else:
                        # The repository changeset_revision is no longer installable, so see if there's been an update.
                        changeset_revision = metadata_util.get_next_downloadable_changeset_revision(
                            self.app, repository, rd_changeset_revision
                        )
                        if changeset_revision != rd_changeset_revision:
                            repository_metadata = (
                                metadata_util.get_repository_metadata_by_repository_id_changeset_revision(
                                    self.app, repository_id, changeset_revision
                                )
                            )
                        if repository_metadata:
                            new_key_rd_dict = {
                                key: [
                                    rd_toolshed,
                                    rd_name,
                                    rd_owner,
                                    repository_metadata.changeset_revision,
                                    rd_prior_installation_required,
                                    rd_only_if_compiling_contained_td,
                                ]
                            }
                            # We have the updated changeset revision.
                            updated_key_rd_dicts.append(new_key_rd_dict)
                        else:
                            repository_components_tuple = container_util.get_components_from_key(key)
                            components_list = tool_shed.util.repository_util.extract_components_from_tuple(
                                repository_components_tuple
                            )
                            (
                                toolshed,
                                repository_name,
                                repository_owner,
                                repository_changeset_revision,
                            ) = components_list[0:4]
                            # For backward compatibility to the 12/20/12 Galaxy release.
                            if len(components_list) in (4, 5):
                                rd_only_if_compiling_contained_td = "False"
                            log.debug(
                                "The revision %s defined for repository %s owned by %s is invalid, so repository "
                                "dependencies defined for repository %s will be ignored.",
                                rd_changeset_revision,
                                rd_name,
                                rd_owner,
                                repository_name,
                            )
                else:
                    repository_components_tuple = container_util.get_components_from_key(key)
                    components_list = tool_shed.util.repository_util.extract_components_from_tuple(
                        repository_components_tuple
                    )
                    toolshed, repository_name, repository_owner, repository_changeset_revision = components_list[0:4]
                    log.debug(
                        "The revision %s defined for repository %s owned by %s is invalid, "
                        "so repository dependencies defined for repository %s will be ignored.",
                        rd_changeset_revision,
                        rd_name,
                        rd_owner,
                        repository_name,
                    )
        return updated_key_rd_dicts

    def handle_circular_repository_dependency(self, repository_key: str, repository_dependency):
        all_repository_dependencies_root_key = self.all_repository_dependencies["root_key"]
        repository_dependency_as_key = self.get_repository_dependency_as_key(repository_dependency)
        self.update_circular_repository_dependencies(
            repository_key, repository_dependency, self.all_repository_dependencies[repository_dependency_as_key]
        )
        if all_repository_dependencies_root_key != repository_dependency_as_key:
            self.all_repository_dependencies[repository_key] = [repository_dependency]

    def handle_current_repository_dependency(self, current_repository_key: str):
        current_repository_key_rd_dicts = []
        for rd in self.all_repository_dependencies[current_repository_key]:
            rd_copy = [str(item) for item in rd]
            new_key_rd_dict = {current_repository_key: rd_copy}
            current_repository_key_rd_dicts.append(new_key_rd_dict)
        if current_repository_key_rd_dicts:
            self.handle_key_rd_dicts_for_repository(current_repository_key, current_repository_key_rd_dicts)
            return self.get_repository_dependencies_for_changeset_revision()

    def handle_key_rd_dicts_for_repository(
        self, current_repository_key, repository_key_rd_dicts: List[Dict[str, List[str]]]
    ):
        key_rd_dict = repository_key_rd_dicts.pop(0)
        repository_dependency = key_rd_dict[current_repository_key]
        (
            toolshed,
            name,
            owner,
            changeset_revision,
            prior_installation_required,
            only_if_compiling_contained_td,
        ) = common_util.parse_repository_dependency_tuple(repository_dependency)
        if suc.tool_shed_is_this_tool_shed(toolshed, trans=self.trans):
            required_repository = get_repository_by_name_and_owner(self.app.model.context, name, owner)
            self.repository = required_repository
            repository_id = self.app.security.encode_id(required_repository.id)
            required_repository_metadata = metadata_util.get_repository_metadata_by_repository_id_changeset_revision(
                self.app, repository_id, changeset_revision
            )
            self.repository_metadata = required_repository_metadata
            if required_repository_metadata:
                # The required_repository_metadata changeset_revision is installable.
                required_metadata = required_repository_metadata.metadata
                if required_metadata:
                    for current_repository_key_rd_dict in repository_key_rd_dicts:
                        if not self.in_key_rd_dicts(current_repository_key_rd_dict, self.key_rd_dicts_to_be_processed):
                            # Add the current repository_dependency into self.key_rd_dicts_to_be_processed.
                            self.key_rd_dicts_to_be_processed.append(current_repository_key_rd_dict)
            if not self.in_key_rd_dicts(key_rd_dict, self.handled_key_rd_dicts):
                # Add the current repository_dependency into self.handled_key_rd_dicts.
                self.handled_key_rd_dicts.append(key_rd_dict)
            if self.in_key_rd_dicts(key_rd_dict, self.key_rd_dicts_to_be_processed):
                # Remove the current repository from self.key_rd_dicts_to_be_processed.
                self.key_rd_dicts_to_be_processed = self.remove_from_key_rd_dicts(
                    key_rd_dict, self.key_rd_dicts_to_be_processed
                )
        else:
            # The repository is in a different tool shed, so build an url and send a request.
            error_message = "Repository dependencies are currently supported only within the same Tool Shed.  "
            error_message += "Ignoring repository dependency definition for tool shed "
            error_message += f"{toolshed}, name {name}, owner {owner}, changeset revision {changeset_revision}"
            log.debug(error_message)

    def handle_next_repository_dependency(self):
        next_repository_key_rd_dict = self.key_rd_dicts_to_be_processed.pop(0)
        next_repository_key_rd_dicts = [next_repository_key_rd_dict]
        next_repository_key = next(iter(next_repository_key_rd_dict))
        self.handle_key_rd_dicts_for_repository(next_repository_key, next_repository_key_rd_dicts)
        return self.get_repository_dependencies_for_changeset_revision()

    def in_all_repository_dependencies(self, repository_key, repository_dependency):
        """
        Return True if { repository_key : repository_dependency } is in self.all_repository_dependencies.
        """
        for key, val in self.all_repository_dependencies.items():
            if key != repository_key:
                continue
            if repository_dependency in val:
                return True
        return False

    def in_circular_repository_dependencies(self, repository_key_rd_dict):
        """
        Return True if any combination of a circular dependency tuple is the key : value pair defined
        in the received repository_key_rd_dict.  This means that each circular dependency tuple is converted
        into the key : value pair for comparison.
        """
        for tup in self.circular_repository_dependencies:
            rd_0, rd_1 = tup
            rd_0_as_key = self.get_repository_dependency_as_key(rd_0)
            rd_1_as_key = self.get_repository_dependency_as_key(rd_1)
            if rd_0_as_key in repository_key_rd_dict and repository_key_rd_dict[rd_0_as_key] == rd_1:
                return True
            if rd_1_as_key in repository_key_rd_dict and repository_key_rd_dict[rd_1_as_key] == rd_0:
                return True
        return False

    def in_key_rd_dicts(self, key_rd_dict: Dict[str, List[str]], key_rd_dicts: List[Dict[str, List[str]]]):
        """Return True if key_rd_dict is contained in the list of key_rd_dicts."""
        k = next(iter(key_rd_dict))
        v = key_rd_dict[k]
        for key_rd_dict in key_rd_dicts:
            for key, val in key_rd_dict.items():
                if key == k and val == v:
                    return True
        return False

    def initialize_all_repository_dependencies(self, current_repository_key: str, repository_dependencies_dict):
        """Initialize the self.all_repository_dependencies dictionary."""
        # It's safe to assume that current_repository_key in this case will have a value.
        self.all_repository_dependencies["root_key"] = current_repository_key
        self.all_repository_dependencies[current_repository_key] = []
        # Store the value of the 'description' key only once, the first time through this recursive method.
        description = repository_dependencies_dict.get("description", None)
        self.all_repository_dependencies["description"] = description

    def is_circular_repository_dependency(self, repository_key, repository_dependency):
        """
        Return True if the received repository_dependency is a key in self.all_repository_dependencies
        whose list of repository dependencies includes the received repository_key.
        """
        repository_dependency_as_key = self.get_repository_dependency_as_key(repository_dependency)
        repository_key_as_repository_dependency = repository_key.split(container_util.STRSEP)
        for key, val in self.all_repository_dependencies.items():
            if key != repository_dependency_as_key:
                continue
            if repository_key_as_repository_dependency in val:
                return True
        return False

    def populate_repository_dependency_objects_for_processing(
        self, current_repository_key: str, repository_dependencies_dict
    ):
        """
        The process that discovers all repository dependencies for a specified repository's changeset
        revision uses this method to populate the following items for the current processing loop:
        filtered_current_repository_key_rd_dicts, self.key_rd_dicts_to_be_processed,
        self.handled_key_rd_dicts, self.all_repository_dependencies.  Each processing loop may discover
        more repository dependencies, so this method is repeatedly called until all repository
        dependencies have been discovered.
        """
        current_repository_key_rd_dicts: List[Dict[str, Any]] = []
        filtered_current_repository_key_rd_dicts = []
        for rd_tup in repository_dependencies_dict["repository_dependencies"]:
            new_key_rd_dict = {current_repository_key: rd_tup}
            current_repository_key_rd_dicts.append(new_key_rd_dict)
        if current_repository_key_rd_dicts and current_repository_key:
            # Remove all repository dependencies that point to a revision within its own repository.
            current_repository_key_rd_dicts = self.remove_repository_dependency_reference_to_self(
                current_repository_key_rd_dicts
            )
        current_repository_key_rd_dicts = self.get_updated_changeset_revisions_for_repository_dependencies(
            current_repository_key_rd_dicts
        )
        for key_rd_dict in current_repository_key_rd_dicts:
            # Filter out repository dependencies that are required only if compiling the dependent
            # repository's tool dependency.
            # TODO: this temporary work-around should be removed when the underlying framework
            # support for handling only_if_compiling_contained_td-flagged repositories is completed.
            key_rd_dict = self.filter_only_if_compiling_contained_td(key_rd_dict)
            if key_rd_dict:
                is_circular = False
                in_handled_key_rd_dicts = self.in_key_rd_dicts(key_rd_dict, self.handled_key_rd_dicts)
                in_key_rd_dicts_to_be_processed = self.in_key_rd_dicts(key_rd_dict, self.key_rd_dicts_to_be_processed)
                if not in_handled_key_rd_dicts and not in_key_rd_dicts_to_be_processed:
                    filtered_current_repository_key_rd_dicts.append(key_rd_dict)
                    repository_dependency = key_rd_dict[current_repository_key]
                    if current_repository_key in self.all_repository_dependencies:
                        # Add all repository dependencies for the current repository into its entry
                        # in self.all_repository_dependencies.
                        all_repository_dependencies_val = self.all_repository_dependencies[current_repository_key]
                        if repository_dependency not in all_repository_dependencies_val:
                            all_repository_dependencies_val.append(repository_dependency)
                            self.all_repository_dependencies[current_repository_key] = all_repository_dependencies_val
                    elif not self.in_all_repository_dependencies(current_repository_key, repository_dependency):
                        # Handle circular repository dependencies.
                        if self.is_circular_repository_dependency(current_repository_key, repository_dependency):
                            is_circular = True
                            self.handle_circular_repository_dependency(current_repository_key, repository_dependency)
                        else:
                            self.all_repository_dependencies[current_repository_key] = [repository_dependency]
                    if not is_circular and self.can_add_to_key_rd_dicts(key_rd_dict, self.key_rd_dicts_to_be_processed):
                        new_key_rd_dict = {current_repository_key: repository_dependency}
                        self.key_rd_dicts_to_be_processed.append(new_key_rd_dict)
        return filtered_current_repository_key_rd_dicts

    def prune_invalid_repository_dependencies(self, repository_dependencies):
        """
        Eliminate all invalid entries in the received repository_dependencies dictionary.  An entry
        is invalid if the value_list of the key/value pair is empty.  This occurs when an invalid
        combination of tool shed, name , owner, changeset_revision is used and a repository_metadata
        record is not found.
        """
        valid_repository_dependencies = {}
        description = repository_dependencies.get("description", None)
        root_key = repository_dependencies.get("root_key", None)
        if root_key is None:
            return valid_repository_dependencies
        for key, value in repository_dependencies.items():
            if key in ["description", "root_key"]:
                continue
            if value:
                valid_repository_dependencies[key] = value
        if valid_repository_dependencies:
            valid_repository_dependencies["description"] = description
            valid_repository_dependencies["root_key"] = root_key
        return valid_repository_dependencies

    def remove_from_key_rd_dicts(self, key_rd_dict: Dict[str, List[str]], key_rd_dicts: List[Dict[str, List[str]]]):
        """Eliminate the key_rd_dict from the list of key_rd_dicts if it is contained in the list."""
        k = next(iter(key_rd_dict))
        v = key_rd_dict[k]
        clean_key_rd_dicts: List[Dict[str, List[str]]] = []
        for krd_dict in key_rd_dicts:
            key = next(iter(krd_dict))
            val = krd_dict[key]
            if key == k and val == v:
                continue
            clean_key_rd_dicts.append(krd_dict)
        return clean_key_rd_dicts

    def remove_repository_dependency_reference_to_self(self, key_rd_dicts: List[Dict[str, Any]]):
        """Remove all repository dependencies that point to a revision within its own repository."""
        clean_key_rd_dicts: List[Dict[str, Any]] = []
        key = next(iter(key_rd_dicts[0]))
        repository_tup = key.split(container_util.STRSEP)
        (
            rd_toolshed,
            rd_name,
            rd_owner,
            rd_changeset_revision,
            rd_prior_installation_required,
            rd_only_if_compiling_contained_td,
        ) = common_util.parse_repository_dependency_tuple(repository_tup)
        cleaned_rd_toolshed = common_util.remove_protocol_from_tool_shed_url(rd_toolshed)
        for key_rd_dict in key_rd_dicts:
            k = next(iter(key_rd_dict))
            repository_dependency = key_rd_dict[k]
            (
                toolshed,
                name,
                owner,
                changeset_revision,
                prior_installation_required,
                only_if_compiling_contained_td,
            ) = common_util.parse_repository_dependency_tuple(repository_dependency)
            cleaned_toolshed = common_util.remove_protocol_from_tool_shed_url(toolshed)
            if cleaned_rd_toolshed == cleaned_toolshed and rd_name == name and rd_owner == owner:
                debug_msg = f"Removing repository dependency for repository {name} owned by {owner} "
                debug_msg += "since it refers to a revision within itself."
                log.debug(debug_msg)
            else:
                new_key_rd_dict = {key: repository_dependency}
                clean_key_rd_dicts.append(new_key_rd_dict)
        return clean_key_rd_dicts

    def update_circular_repository_dependencies(
        self, repository_key: str, repository_dependency, repository_dependencies
    ):
        repository_key_as_repository_dependency = repository_key.split(container_util.STRSEP)
        if repository_key_as_repository_dependency in repository_dependencies:
            found = False
            for tup in self.circular_repository_dependencies:
                if repository_dependency in tup and repository_key_as_repository_dependency in tup:
                    # The circular dependency has already been included.
                    found = True
            if not found:
                new_circular_tup = (repository_dependency, repository_key_as_repository_dependency)
                self.circular_repository_dependencies.append(new_circular_tup)



from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

column_repository_name = "column_maker_0150"
column_repository_description = "Add column"
column_repository_long_description = "Compute an expression on every row"

convert_repository_name = "convert_chars_0150"
convert_repository_description = "Convert delimiters"
convert_repository_long_description = "Convert delimiters to tab"

category_name = "Test 0150 Simple Prior Installation"
category_description = "Test 0150 Simple Prior Installation"

"""
Create column_maker and convert_chars.

Column maker repository dependency:
<repository toolshed="self.url" name="convert_chars" owner="test" changeset_revision="c3041382815c" prior_installation_required="True" />

Verify display.

Galaxy side:

Install column_maker.
Verify that convert_chars was installed first, contrary to the ordering that would be present without prior_installation_required.
"""


class TestSimplePriorInstallation(ShedTwillTestCase):
    """Test features related to datatype converters."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_convert_repository(self):
        """Create and populate convert_chars_0150."""
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=convert_repository_name,
            description=convert_repository_description,
            long_description=convert_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "convert_chars/convert_chars.tar",
            commit_message="Uploaded convert_chars tarball.",
        )

    def test_0010_create_column_repository(self):
        """Create and populate convert_chars_0150."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=column_repository_name,
            description=column_repository_description,
            long_description=column_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "column_maker/column_maker.tar",
            commit_message="Uploaded column_maker tarball.",
        )

    def test_0015_create_repository_dependency(self):
        """Create a repository dependency specifying convert_chars.

        Column maker repository dependency:
            <repository toolshed="self.url" name="convert_chars" owner="test" changeset_revision="<tip>" prior_installation_required="True" />
        """
        column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
        convert_repository = self._get_repository_by_name_and_owner(convert_repository_name, common.test_user_1_name)
        dependency_xml_path = self.generate_temp_path("test_0150", additional_paths=["column"])
        convert_tuple = (
            self.url,
            convert_repository.name,
            convert_repository.owner,
            self.get_repository_tip(convert_repository),
        )
        self.create_repository_dependency(
            repository=column_repository,
            repository_tuples=[convert_tuple],
            filepath=dependency_xml_path,
            prior_installation_required=True,
        )

    def test_0020_verify_repository_dependency(self):
        """Verify that the previously generated repositiory dependency displays correctly."""
        column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
        convert_repository = self._get_repository_by_name_and_owner(convert_repository_name, common.test_user_1_name)
        self.check_repository_dependency(
            repository=column_repository,
            depends_on_repository=convert_repository,
            depends_on_changeset_revision=None,
            changeset_revision=None,
        )



import os

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase


class TestUninstallingAndReinstallingRepositories(ShedTwillTestCase):
    """Test uninstalling and reinstalling a repository with tool dependencies."""

    requires_galaxy = True

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_ensure_repositories_and_categories_exist(self):
        """Create the 0010 category and upload the freebayes repository to the tool shed, if necessary."""
        category = self.create_category(
            name="Test 0010 Repository With Tool Dependencies",
            description="Tests for a repository with tool dependencies.",
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name="freebayes_0010",
            description="Galaxy's freebayes tool",
            long_description="Long description of Galaxy's freebayes tool",
            owner=common.test_user_1_name,
            category=category,
        )
        if self.repository_is_new(repository):
            strings_displayed = [
                "Metadata may have been defined",
                "This file requires an entry",
                "tool_data_table_conf",
            ]
            self.add_file_to_repository(repository, "freebayes/freebayes.xml", strings_displayed=strings_displayed)
            strings_displayed = ["Upload a file named <b>sam_fa_indices.loc.sample"]
            self.add_file_to_repository(
                repository, "freebayes/tool_data_table_conf.xml.sample", strings_displayed=strings_displayed
            )
            self.add_file_to_repository(repository, "freebayes/sam_fa_indices.loc.sample")
            target = os.path.join("freebayes", "malformed_tool_dependencies", "tool_dependencies.xml")
            self.add_file_to_repository(
                repository, target, strings_displayed=["Exception attempting to parse", "invalid element name"]
            )
            target = os.path.join("freebayes", "invalid_tool_dependencies", "tool_dependencies.xml")
            strings_displayed = [
                "The settings for <b>name</b>, <b>version</b> and <b>type</b> from a contained tool configuration"
            ]
            self.add_file_to_repository(repository, target, strings_displayed=strings_displayed)
            target = os.path.join("freebayes", "tool_dependencies.xml")
            self.add_file_to_repository(repository, target)

    def test_0010_install_freebayes_repository(self):
        """Install the freebayes repository into the Galaxy instance."""
        self._install_repository(
            "freebayes_0010",
            common.test_user_1_name,
            "Test 0010 Repository With Tool Dependencies",
            new_tool_panel_section_label="test_1210",
        )
        self._assert_has_installed_repos_with_names("freebayes_0010")

    def test_0015_uninstall_freebayes_repository(self):
        """Uninstall the freebayes repository."""
        installed_repository = self._get_installed_repository_by_name_owner("freebayes_0010", common.test_user_1_name)
        self._uninstall_repository(installed_repository)
        self._assert_has_no_installed_repos_with_names("freebayes_0010")

    def test_0020_reinstall_freebayes_repository(self):
        """Reinstall the freebayes repository."""
        installed_repository = self._get_installed_repository_by_name_owner("freebayes_0010", common.test_user_1_name)
        self.reinstall_repository_api(installed_repository)
        self._assert_has_installed_repos_with_names("freebayes_0010")
        self._assert_has_valid_tool_with_name("FreeBayes")
        self._assert_repo_has_tool_with_id(installed_repository, "freebayes")

    def test_0025_deactivate_freebayes_repository(self):
        """Deactivate the freebayes repository without removing it from disk."""
        installed_repository = self._get_installed_repository_by_name_owner("freebayes_0010", common.test_user_1_name)
        self.deactivate_repository(installed_repository)
        self._assert_has_no_installed_repos_with_names("freebayes_0010")

    def test_0030_reactivate_freebayes_repository(self):
        """Reactivate the freebayes repository and verify that it now shows up in the list of installed repositories."""
        installed_repository = self._get_installed_repository_by_name_owner("freebayes_0010", common.test_user_1_name)
        self.reactivate_repository(installed_repository)
        self._assert_has_installed_repos_with_names("freebayes_0010")
        self._assert_has_valid_tool_with_name("FreeBayes")
        self._assert_repo_has_tool_with_id(installed_repository, "freebayes")



import logging

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

repository_name = "htseq_count_0140"
repository_description = "Converter: BED to GFF"
repository_long_description = "Convert bed to gff"

category_name = "Test 0140 Tool Help Images"
category_description = "Test 0140 Tool Help Images"

# 1) Create and populate the htseq_count_0140 repository.
# 2) Visit the manage_repository page, then the tool page, and look for the image string
# similar to the following string where the encoded repository_id is previously determined:
# src="/repository/static/images/<id>/count_modes.png"


class TestToolHelpImages(ShedTwillTestCase):
    """Test features related to tool help images."""

    requires_galaxy = True

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_htseq_count_repository(self):
        """Create and populate htseq_count_0140.

        We are at step 1 - Create and populate the htseq_count_0140 repository.
        Create the htseq_count_0140 repository and upload the tarball.
        """
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        # Create a repository named htseq_count_0140 owned by user1.
        repository = self.get_or_create_repository(
            name=repository_name,
            description=repository_description,
            long_description=repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            # Upload htseq_count.tar to the repository if it hasn't already been populated.
            self.commit_tar_to_repository(
                repository,
                "htseq_count/htseq_count.tar",
                commit_message="Uploaded htseq_count.tar.",
            )

    def test_0010_load_tool_page(self):
        """Load the tool page and check for the image URL.

        This is a duplicate of test method _0010 in test_0140_tool_help_images.
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        # Get the repository tip.
        changeset_revision = self.get_repository_tip(repository)
        # Generate the image path.
        image_path = f'src="/repository/static/images/{repository.id}/count_modes.png"'
        # The repository uploaded in this test should only have one metadata revision, with one tool defined, which
        # should be the tool that contains a link to the image.
        repository_metadata = self._db_repository(repository).metadata_revisions[0].metadata
        tool_path = repository_metadata["tools"][0]["tool_config"]
        # V2 is not going to have this page right? So... do we need this test at all or that route? Likely not?
        if self._browser.is_twill and not self.is_v2:
            self.load_display_tool_page(
                repository, tool_path, changeset_revision, strings_displayed=[image_path], strings_not_displayed=[]
            )



import logging
from typing import TYPE_CHECKING

from tool_shed.util import (
    hg_util,
    metadata_util,
    repository_util,
)

if TYPE_CHECKING:
    from tool_shed.structured_app import ToolShedApp

log = logging.getLogger(__name__)


class ToolVersionManager:
    def __init__(self, app: "ToolShedApp"):
        self.app = app

    def get_version_lineage_for_tool(self, repository_id, repository_metadata, guid):
        """
        Return the tool version lineage chain in descendant order for the received
        guid contained in the received repsitory_metadata.tool_versions.  This function
        is called only from the Tool Shed.
        """
        repository = repository_util.get_repository_by_id(self.app, repository_id)
        repo = repository.hg_repo
        # Initialize the tool lineage
        version_lineage = [guid]
        # Get all ancestor guids of the received guid.
        current_child_guid = guid
        for changeset in hg_util.reversed_upper_bounded_changelog(repo, repository_metadata.changeset_revision):
            ctx = repo[changeset]
            rm = metadata_util.get_repository_metadata_by_changeset_revision(self.app, repository_id, str(ctx))
            if rm:
                parent_guid = rm.tool_versions.get(current_child_guid, None)
                if parent_guid:
                    version_lineage.append(parent_guid)
                    current_child_guid = parent_guid
        # Get all descendant guids of the received guid.
        current_parent_guid = guid
        for changeset in hg_util.reversed_lower_upper_bounded_changelog(
            repo, repository_metadata.changeset_revision, repository.tip()
        ):
            ctx = repo[changeset]
            rm = metadata_util.get_repository_metadata_by_changeset_revision(self.app, repository_id, str(ctx))
            if rm:
                tool_versions = rm.tool_versions
                for child_guid, parent_guid in tool_versions.items():
                    if parent_guid == current_parent_guid:
                        version_lineage.insert(0, child_guid)
                        current_parent_guid = child_guid
                        break
        return version_lineage



import time
from typing import List

from playwright.sync_api import (
    expect,
    Locator,
    Page,
)

from .browser import (
    FormValueType,
    ShedBrowser,
)


class Locators:
    toolbar_login = ".toolbar-login"
    toolbar_logout = ".toolbar-logout"
    login_submit_button = '[name="login_button"]'
    register_link = ".register-link"


class PlaywrightShedBrowser(ShedBrowser):
    _page: Page

    def __init__(self, page: Page):
        self._page = page

    def visit_url(self, url: str, allowed_codes: List[int]) -> str:
        try:
            response = self._page.goto(url)
        except Exception as e:
            if "Navigation interrupted by another one" in str(e):
                # I believe redirect on the target page interfering with
                # this thread's test.
                time.sleep(0.25)
                response = self._page.goto(url)
            else:
                raise
        assert response is not None
        return_code = response.status
        assert return_code in allowed_codes, "Invalid HTTP return code {}, allowed codes: {}".format(
            return_code,
            ", ".join(str(code) for code in allowed_codes),
        )
        return response.url

    def page_content(self) -> str:
        self._page.wait_for_load_state("networkidle")
        return self._page.content()

    def check_page_for_string(self, patt: str) -> None:
        """Looks for 'patt' in the current browser page"""
        patt = patt.replace("<b>", "").replace("</b>", "")
        expect(self._page.locator("body")).to_contain_text(patt)

    def check_string_not_in_page(self, patt: str) -> None:
        patt = patt.replace("<b>", "").replace("</b>", "")
        expect(self._page.locator("body")).not_to_contain_text(patt)

    def xcheck_page_for_string(self, patt: str) -> None:
        page = self.page_content()
        if page.find(patt) == -1:
            fname = self.write_temp_file(page)
            errmsg = f"no match to '{patt}'\npage content written to '{fname}'\npage: [[{page}]]"
            raise AssertionError(errmsg)

    def xcheck_string_not_in_page(self, patt: str) -> None:
        page = self.page_content()
        if page.find(patt) != -1:
            fname = self.write_temp_file(page)
            errmsg = f"string ({patt}) incorrectly displayed in page.\npage content written to '{fname}'"
            raise AssertionError(errmsg)

    def write_temp_file(self, content, suffix=".html"):
        import tempfile

        from galaxy.util import smart_str

        with tempfile.NamedTemporaryFile(suffix=suffix, prefix="twilltestcase-", delete=False) as fh:
            fh.write(smart_str(content))
        return fh.name

    def show_forms(self) -> Locator:
        """Shows form, helpful for debugging new tests"""
        return self._page.locator("form")

    def submit_form_with_name(self, form_name: str, button="runtool_btn", **kwd):
        form = self._form_with_name(form_name)
        self._submit_form(form, button, **kwd)

    def submit_form(self, form_no=-1, button="runtool_btn", form=None, **kwd):
        """Populates and submits a form from the keyword arguments."""
        # An HTMLForm contains a sequence of Controls.  Supported control classes are:
        # TextControl, FileControl, ListControl, RadioControl, CheckboxControl, SelectControl,
        # SubmitControl, ImageControl
        if form is None:
            try:
                form = self.show_forms().nth(form_no)
            except IndexError:
                raise ValueError("No form to submit found")
        self._submit_form(form, button, **kwd)

    def _submit_form(self, form: Locator, button="runtool_btn", **kwd):
        for control_name, control_value in kwd.items():
            self._fill_form_value(form, control_name, control_value)
        input = self._page.locator(f"[name='{button}']")
        if input.count():
            input.click()
        else:
            submit_input = form.locator("input[type=submit]")
            submit_input.click()
        time.sleep(0.25)
        # tc.submit(button)

    def _form_with_name(self, name: str) -> Locator:
        forms = self.show_forms()
        count = forms.count()
        for i in range(count):
            nth_form = self.show_forms().nth(i)
            if nth_form.get_attribute("name") == name:
                return nth_form
        raise KeyError(f"No form with name [{name}]")

    def fill_form_value(self, form_name: str, control_name: str, value: FormValueType):
        form: Locator = self._form_with_name(form_name)
        self._fill_form_value(form, control_name, value)

    def _fill_form_value(self, form: Locator, control_name: str, value: FormValueType):
        input_i = form.locator(f"input[name='{control_name}']")
        input_t = form.locator(f"textarea[name='{control_name}']")
        input_s = form.locator(f"select[name='{control_name}']")
        if input_i.count():
            if control_name in ["redirect"]:
                input_i.input_value = value  # type:ignore[assignment, unused-ignore]
            else:
                if isinstance(value, bool):
                    if value and not input_i.is_checked():
                        input_i.check()
                    elif not value and input_i.is_checked():
                        input_i.uncheck()
                else:
                    input_i.fill(value)
        if input_t.count():
            input_t.fill(value)  # type:ignore[arg-type, unused-ignore]
        if input_s.count():
            input_s.select_option(value)  # type:ignore[arg-type, unused-ignore]

    def edit_repository_categories(self, categories_to_add: List[str], categories_to_remove: List[str]) -> None:
        multi_select = "form[name='categories'] select[name='category_id']"
        select_locator = self._page.locator(multi_select)
        select_locator.evaluate("node => node.selectedOptions = []")
        select_locator.select_option(label=categories_to_add)
        self.submit_form_with_name("categories", "manage_categories_button")

        select_locator.evaluate("node => node.selectedOptions = []")
        select_locator.select_option(label=categories_to_remove)
        self.submit_form_with_name("categories", "manage_categories_button")

    def grant_users_access(self, usernames: List[str]):
        multi_select = "form[name='user_access'] select[name='allow_push']"
        select_locator = self._page.locator(multi_select)
        select_locator.evaluate("node => node.selectedOptions = []")
        select_locator.select_option(label=usernames)
        self.submit_form_with_name("user_access", "user_access_button")

    @property
    def is_twill(self) -> bool:
        return False

    def logout_if_logged_in(self, assert_logged_out=True):
        self._page.wait_for_selector(f"{Locators.toolbar_login}, {Locators.toolbar_logout}")
        logout_locator = self._page.locator(Locators.toolbar_logout)
        if logout_locator.is_visible():
            logout_locator.click()
            self._page.wait_for_load_state("networkidle")
        if assert_logged_out:
            self.expect_not_logged_in()

    def explicit_logout(self):
        self._page.wait_for_selector(Locators.toolbar_logout)
        logout_locator = self._page.locator(Locators.toolbar_logout)
        logout_locator.click()
        self._page.wait_for_load_state("networkidle")
        expect(self._page.locator(Locators.toolbar_logout)).not_to_be_visible()

    def expect_not_logged_in(self):
        expect(self._page.locator(Locators.toolbar_logout)).not_to_be_visible()

    def expect_logged_in(self):
        expect(self._page.locator(Locators.toolbar_logout)).to_be_visible()



import logging

from galaxy.web.form_builder import SelectField
from tool_shed.util import (
    hg_util,
    metadata_util,
)

log = logging.getLogger(__name__)


def build_changeset_revision_select_field(
    trans,
    repository,
    selected_value=None,
    add_id_to_name=True,
    downloadable=False,
):
    """
    Build a SelectField whose options are the changeset_rev strings of certain revisions of the
    received repository.
    """
    options = []
    changeset_tups = []
    refresh_on_change_values = []
    if downloadable:
        # Restrict the options to downloadable revisions.
        repository_metadata_revisions = repository.downloadable_revisions
    else:
        # Restrict the options to all revisions that have associated metadata.
        repository_metadata_revisions = repository.metadata_revisions
    for repository_metadata in repository_metadata_revisions:
        rev, label, changeset_revision = hg_util.get_rev_label_changeset_revision_from_repository_metadata(
            trans.app, repository_metadata, repository=repository, include_date=True, include_hash=False
        )
        changeset_tups.append((rev, label, changeset_revision))
        refresh_on_change_values.append(changeset_revision)
    # Sort options by the revision label.  Even though the downloadable_revisions query sorts by update_time,
    # the changeset revisions may not be sorted correctly because setting metadata over time will reset update_time.
    for changeset_tup in sorted(changeset_tups):
        # Display the latest revision first.
        options.insert(0, (changeset_tup[1], changeset_tup[2]))
    if add_id_to_name:
        name = f"changeset_revision_{repository.id}"
    else:
        name = "changeset_revision"
    select_field = SelectField(name=name, refresh_on_change=True)
    for option_tup in options:
        selected = selected_value and option_tup[1] == selected_value
        select_field.add_option(option_tup[0], option_tup[1], selected=selected)
    return select_field


def filter_by_latest_downloadable_changeset_revision_that_has_missing_tool_test_components(trans, repository):
    """
    Inspect the latest downloadable changeset revision for the received repository to see if it
    includes tools that are either missing functional tests or functional test data.  If the
    changset revision includes tools but is missing tool test components, return the changeset
    revision hash.  This will filter out repositories of type repository_suite_definition and
    tool_dependency_definition.
    """
    repository_metadata = get_latest_downloadable_repository_metadata_if_it_includes_tools(trans, repository)
    if repository_metadata is not None and repository_metadata.missing_test_components:
        return repository_metadata.changeset_revision
    return None


def filter_by_latest_metadata_changeset_revision_that_has_invalid_tools(trans, repository):
    """
    Inspect the latest changeset revision with associated metadata for the received repository
    to see if it has invalid tools.  This will filter out repositories of type repository_suite_definition
    and tool_dependency_definition.
    """
    repository_metadata = get_latest_repository_metadata_if_it_includes_invalid_tools(trans, repository)
    if repository_metadata is not None:
        return repository_metadata.changeset_revision
    return None


def get_latest_downloadable_repository_metadata(trans, repository):
    """
    Return the latest downloadable repository_metadata record for the received repository.  This will
    return repositories of type unrestricted as well as types repository_suite_definition and
    tool_dependency_definition.
    """
    encoded_repository_id = trans.security.encode_id(repository.id)
    repo = repository.hg_repo
    tip_ctx = str(repo[repo.changelog.tip()])
    repository_metadata = None
    try:
        repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
            trans.app, encoded_repository_id, tip_ctx
        )
        if repository_metadata is not None and repository_metadata.downloadable:
            return repository_metadata
        return None
    except Exception:
        latest_downloadable_revision = metadata_util.get_previous_metadata_changeset_revision(
            trans.app, repository, tip_ctx, downloadable=True
        )
        if latest_downloadable_revision == hg_util.INITIAL_CHANGELOG_HASH:
            return None
        repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
            trans.app, encoded_repository_id, latest_downloadable_revision
        )
        if repository_metadata is not None and repository_metadata.downloadable:
            return repository_metadata
        return None


def get_latest_downloadable_repository_metadata_if_it_includes_tools(trans, repository):
    """
    Return the latest downloadable repository_metadata record for the received repository if its
    includes_tools attribute is True.  This will filter out repositories of type repository_suite_definition
    and tool_dependency_definition.
    """
    repository_metadata = get_latest_downloadable_repository_metadata(trans, repository)
    if repository_metadata is not None and repository_metadata.includes_tools:
        return repository_metadata
    return None


def get_latest_repository_metadata(trans, repository):
    """
    Return the latest repository_metadata record for the received repository if it exists.  This will
    return repositories of type unrestricted as well as types repository_suite_definition and
    tool_dependency_definition.
    """
    encoded_repository_id = trans.security.encode_id(repository.id)
    repo = repository.hg_repo
    tip_ctx = str(repo[repo.changelog.tip()])
    try:
        repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
            trans.app, encoded_repository_id, tip_ctx
        )
        return repository_metadata
    except Exception:
        latest_downloadable_revision = metadata_util.get_previous_metadata_changeset_revision(
            trans.app, repository, tip_ctx, downloadable=False
        )
        if latest_downloadable_revision == hg_util.INITIAL_CHANGELOG_HASH:
            return None
        repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
            trans.app, encoded_repository_id, latest_downloadable_revision
        )
        return repository_metadata


def get_latest_repository_metadata_if_it_includes_invalid_tools(trans, repository):
    """
    Return the latest repository_metadata record for the received repository that contains invalid
    tools if one exists.  This will filter out repositories of type repository_suite_definition and
    tool_dependency_definition.
    """
    repository_metadata = get_latest_repository_metadata(trans, repository)
    if repository_metadata is not None:
        metadata = repository_metadata.metadata
        if metadata is not None and "invalid_tools" in metadata:
            return repository_metadata
    return None



import json
import logging
import os

from mako.template import Template

from galaxy import web
from galaxy.tool_shed.util import basic_util
from galaxy.tool_shed.util.hg_util import (
    get_changectx_for_changeset,
    get_file_context_from_ctx,
)
from galaxy.tool_shed.util.repository_util import get_repo_info_tuple_contents
from galaxy.tool_shed.util.shed_util_common import set_image_paths
from galaxy.util import (
    rst_to_html,
    unicodify,
    url_get,
)
from galaxy.util.tool_shed.common_util import get_tool_shed_url_from_tool_shed_registry
from tool_shed.util.metadata_util import get_latest_downloadable_changeset_revision

log = logging.getLogger(__name__)


def build_readme_files_dict(app, repository, changeset_revision, metadata, tool_path=None):
    """
    Return a dictionary of valid readme file name <-> readme file content pairs for all readme files defined in the received metadata.  Since the
    received changeset_revision (which is associated with the received metadata) may not be the latest installable changeset revision, the README
    file contents may not be available on disk.  This method is used by both Galaxy and the Tool Shed.
    """
    if app.name == "galaxy":
        can_use_disk_files = True
    else:
        latest_downloadable_changeset_revision = get_latest_downloadable_changeset_revision(app, repository)
        can_use_disk_files = changeset_revision == latest_downloadable_changeset_revision
    readme_files_dict = {}
    if metadata:
        if "readme_files" in metadata:
            for relative_path_to_readme_file in metadata["readme_files"]:
                readme_file_name = os.path.split(relative_path_to_readme_file)[1]
                if can_use_disk_files:
                    if tool_path:
                        full_path_to_readme_file = os.path.abspath(
                            os.path.join(tool_path, relative_path_to_readme_file)
                        )
                    else:
                        full_path_to_readme_file = os.path.abspath(relative_path_to_readme_file)
                    text = None
                    try:
                        with open(full_path_to_readme_file, encoding="utf-8") as f:
                            text = f.read()
                    except Exception:
                        log.exception("Error reading README file '%s' from disk", relative_path_to_readme_file)
                        text = None
                    if text:
                        text_of_reasonable_length = basic_util.size_string(text)
                        if text_of_reasonable_length.find(".. image:: ") >= 0:
                            # Handle image display for README files that are contained in repositories in the tool shed or installed into Galaxy.
                            try:
                                text_of_reasonable_length = set_image_paths(
                                    app,
                                    text_of_reasonable_length,
                                    encoded_repository_id=app.security.encode_id(repository.id),
                                )
                            except Exception:
                                log.exception(
                                    "Exception in build_readme_files_dict, so images may not be properly displayed"
                                )
                        if readme_file_name.endswith(".rst"):
                            text_of_reasonable_length = Template(
                                rst_to_html(text_of_reasonable_length),
                                input_encoding="utf-8",
                                default_filters=["decode.utf8"],
                                encoding_errors="replace",
                            )
                            text_of_reasonable_length = text_of_reasonable_length.render(
                                static_path=web.url_for("/static"), host_url=web.url_for("/", qualified=True)
                            )
                            text_of_reasonable_length = unicodify(text_of_reasonable_length)
                        else:
                            text_of_reasonable_length = basic_util.to_html_string(text_of_reasonable_length)
                        readme_files_dict[readme_file_name] = text_of_reasonable_length
                else:
                    # We must be in the tool shed and have an old changeset_revision, so we need to retrieve the file contents from the repository manifest.
                    repo = repository.hg_repo
                    ctx = get_changectx_for_changeset(repo, changeset_revision)
                    if ctx:
                        fctx = get_file_context_from_ctx(ctx, readme_file_name)
                        if fctx and fctx not in ["DELETED"]:
                            try:
                                text = unicodify(fctx.data())
                                readme_files_dict[readme_file_name] = basic_util.size_string(text)
                            except Exception:
                                log.exception(
                                    "Error reading README file '%s' from repository manifest",
                                    relative_path_to_readme_file,
                                )
    return readme_files_dict


def get_readme_files_dict_for_display(app, tool_shed_url, repo_info_dict):
    """
    Return a dictionary of README files contained in the single repository being installed so they can be displayed on the tool panel section
    selection page.
    """
    name = next(iter(repo_info_dict))
    repo_info_tuple = repo_info_dict[name]
    (
        description,
        repository_clone_url,
        changeset_revision,
        ctx_rev,
        repository_owner,
        repository_dependencies,
        installed_td,
    ) = get_repo_info_tuple_contents(repo_info_tuple)
    # Handle changing HTTP protocols over time.
    tool_shed_url = get_tool_shed_url_from_tool_shed_registry(app, tool_shed_url)
    params = dict(name=name, owner=repository_owner, changeset_revision=changeset_revision)
    pathspec = ["repository", "get_readme_files"]
    raw_text = url_get(
        tool_shed_url, auth=app.tool_shed_registry.url_auth(tool_shed_url), pathspec=pathspec, params=params
    )
    readme_files_dict = json.loads(raw_text)
    return readme_files_dict



from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

column_maker_repository_name = "column_maker_0020"
column_maker_repository_description = "A flexible aligner."
column_maker_repository_long_description = "A flexible aligner and methylation caller for Bisulfite-Seq applications."

emboss_repository_name = "emboss_0020"
emboss_repository_description = "Galaxy wrappers for Emboss version 5.0.0 tools for test 0020"
emboss_repository_long_description = "Galaxy wrappers for Emboss version 5.0.0 tools for test 0020"


class TestToolWithRepositoryDependencies(ShedTwillTestCase):
    """Test installing a repository with repository dependencies."""

    requires_galaxy = True

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_ensure_repositories_and_categories_exist(self):
        """Create the 0020 category and any missing repositories."""
        category = self.create_category(
            name="Test 0020 Basic Repository Dependencies", description="Test 0020 Basic Repository Dependencies"
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        column_maker_repository = self.get_or_create_repository(
            name=column_maker_repository_name,
            description=column_maker_repository_description,
            long_description=column_maker_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(column_maker_repository):
            self.commit_tar_to_repository(
                column_maker_repository,
                "column_maker/column_maker.tar",
                commit_message="Uploaded column_maker tarball.",
            )
            emboss_repository = self.get_or_create_repository(
                name=emboss_repository_name,
                description=emboss_repository_description,
                long_description=emboss_repository_long_description,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                emboss_repository,
                "emboss/emboss.tar",
                commit_message="Uploaded emboss.tar",
            )
            repository_dependencies_path = self.generate_temp_path("test_1020", additional_paths=["emboss", "5"])
            repository_tuple = (
                self.url,
                column_maker_repository.name,
                column_maker_repository.owner,
                self.get_repository_tip(column_maker_repository),
            )
            self.create_repository_dependency(
                repository=emboss_repository,
                repository_tuples=[repository_tuple],
                filepath=repository_dependencies_path,
            )

    def test_0010_browse_tool_shed(self):
        """Browse the available tool sheds in this Galaxy instance and preview the emboss tool."""
        self.browse_tool_shed(url=self.url, strings_displayed=["Test 0020 Basic Repository Dependencies"])
        category = self.populator.get_category_with_name("Test 0020 Basic Repository Dependencies")
        self.browse_category(category, strings_displayed=[emboss_repository_name])
        if not self.is_v2:
            self.preview_repository_in_tool_shed(
                emboss_repository_name,
                common.test_user_1_name,
                strings_displayed=[emboss_repository_name, "Valid tools"],
            )

    def test_0015_install_emboss_repository(self):
        """Install the emboss repository without installing tool dependencies."""
        self._install_repository(
            emboss_repository_name,
            common.test_user_1_name,
            "Test 0020 Basic Repository Dependencies",
            install_tool_dependencies=False,
            new_tool_panel_section_label="test_1020",
        )
        installed_repository = self._get_installed_repository_by_name_owner(
            emboss_repository_name, common.test_user_1_name
        )
        assert self._get_installed_repository_for(
            common.test_user_1, emboss_repository_name, installed_repository.installed_changeset_revision
        )
        self._assert_has_valid_tool_with_name("antigenic")
        self._assert_repo_has_tool_with_id(installed_repository, "EMBOSS: antigenic1")

    def test_0020_verify_installed_repository_metadata(self):
        """Verify that resetting the metadata on an installed repository does not change the metadata."""
        self.verify_installed_repository_metadata_unchanged(emboss_repository_name, common.test_user_1_name)

    def test_0025_deactivate_datatypes_repository(self):
        """Deactivate the emboss_datatypes repository without removing it from disk."""
        installed_repository = self._get_installed_repository_by_name_owner(
            column_maker_repository_name, common.test_user_1_name
        )
        self.deactivate_repository(installed_repository)

    def test_0030_reactivate_datatypes_repository(self):
        """Reactivate the datatypes repository and verify that the datatypes are again present."""
        installed_repository = self._get_installed_repository_by_name_owner(
            column_maker_repository_name, common.test_user_1_name
        )
        self.reactivate_repository(installed_repository)
        # This used to reactive datatype repositories and verify counts...
        # test may be considerably less useful now.



from ..base import common
from ..base.api import skip_if_api_v2
from ..base.twilltestcase import ShedTwillTestCase

emboss_repository_name = "emboss_0050"
emboss_repository_description = "Galaxy's emboss tool"
emboss_repository_long_description = "Long description of Galaxy's emboss tool"

filtering_repository_name = "filtering_0050"
filtering_repository_description = "Galaxy's filtering tool"
filtering_repository_long_description = "Long description of Galaxy's filtering tool"

freebayes_repository_name = "freebayes_0050"
freebayes_repository_description = "Galaxy's freebayes tool"
freebayes_repository_long_description = "Long description of Galaxy's freebayes tool"

column_repository_name = "column_maker_0050"
column_repository_description = "Add column"
column_repository_long_description = "Compute an expression on every row"

convert_repository_name = "convert_chars_0050"
convert_repository_description = "Convert delimiters"
convert_repository_long_description = "Convert delimiters to tab"

bismark_repository_name = "bismark_0050"
bismark_repository_description = "A flexible aligner."
bismark_repository_long_description = "A flexible aligner and methylation caller for Bisulfite-Seq applications."

category_name = "Test 0050 Circular Dependencies 5 Levels"
category_description = "Test circular dependency features"


class TestRepositoryCircularDependenciesToNLevels(ShedTwillTestCase):
    """Verify that the code correctly handles circular dependencies down to n levels."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_convert_repository(self):
        """Create and populate convert_chars_0050."""
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=convert_repository_name,
            description=convert_repository_description,
            long_description=convert_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "convert_chars/convert_chars.tar",
            commit_message="Uploaded convert_chars tarball.",
        )

    def test_0010_create_column_repository(self):
        """Create and populate convert_chars_0050."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=column_repository_name,
            description=column_repository_description,
            long_description=column_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "column_maker/column_maker.tar",
            commit_message="Uploaded column_maker tarball.",
        )

    def test_0015_create_emboss_datatypes_repository(self):
        """noop"""
        pass

    def test_0020_create_emboss_repository(self):
        """Create and populate emboss_0050."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=emboss_repository_name,
            description=emboss_repository_description,
            long_description=emboss_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "emboss/emboss.tar",
            commit_message="Uploaded emboss tarball.",
        )

    def test_0025_create_filtering_repository(self):
        """Create and populate filtering_0050."""
        category = self.create_category(name=category_name, description=category_description)
        filtering_repository = self.get_or_create_repository(
            name=filtering_repository_name,
            description=filtering_repository_description,
            long_description=filtering_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            filtering_repository,
            "filtering/filtering_1.1.0.tar",
            commit_message="Uploaded filtering 1.1.0 tarball.",
        )

    def test_0030_create_freebayes_repository(self):
        """Create and populate freebayes_0050."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=freebayes_repository_name,
            description=freebayes_repository_description,
            long_description=freebayes_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "freebayes/freebayes.tar",
            commit_message="Uploaded freebayes tarball.",
        )

    def test_0035_create_bismark_repository(self):
        """Create and populate bismark_0050."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=bismark_repository_name,
            description=bismark_repository_description,
            long_description=bismark_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.user_populator().setup_bismark_repo(repository, end=1)

    def test_0040_create_and_upload_dependency_definitions(self):
        column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
        convert_repository = self._get_repository_by_name_and_owner(convert_repository_name, common.test_user_1_name)
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        filtering_repository = self._get_repository_by_name_and_owner(
            filtering_repository_name, common.test_user_1_name
        )
        freebayes_repository = self._get_repository_by_name_and_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        bismark_repository = self._get_repository_by_name_and_owner(bismark_repository_name, common.test_user_1_name)
        dependency_xml_path = self.generate_temp_path("test_0050", additional_paths=["freebayes"])
        # convert_chars depends on column_maker
        # column_maker depends on convert_chars
        # emboss depends on emboss_datatypes
        # emboss_datatypes depends on bismark
        # freebayes depends on freebayes, emboss, emboss_datatypes, and column_maker
        # filtering depends on emboss
        column_tuple = (
            self.url,
            column_repository.name,
            column_repository.owner,
            self.get_repository_tip(column_repository),
        )
        convert_tuple = (
            self.url,
            convert_repository.name,
            convert_repository.owner,
            self.get_repository_tip(convert_repository),
        )
        freebayes_tuple = (
            self.url,
            freebayes_repository.name,
            freebayes_repository.owner,
            self.get_repository_tip(freebayes_repository),
        )
        emboss_tuple = (
            self.url,
            emboss_repository.name,
            emboss_repository.owner,
            self.get_repository_tip(emboss_repository),
        )
        bismark_tuple = (
            self.url,
            bismark_repository.name,
            bismark_repository.owner,
            self.get_repository_tip(bismark_repository),
        )
        self.create_repository_dependency(
            repository=convert_repository, repository_tuples=[column_tuple], filepath=dependency_xml_path
        )
        self.create_repository_dependency(
            repository=column_repository, repository_tuples=[convert_tuple], filepath=dependency_xml_path
        )
        self.create_repository_dependency(
            repository=emboss_repository, repository_tuples=[bismark_tuple], filepath=dependency_xml_path
        )
        self.create_repository_dependency(
            repository=freebayes_repository,
            repository_tuples=[freebayes_tuple, emboss_tuple, column_tuple],
            filepath=dependency_xml_path,
        )
        self.create_repository_dependency(
            repository=filtering_repository, repository_tuples=[emboss_tuple], filepath=dependency_xml_path
        )

    def test_0045_verify_repository_dependencies(self):
        """Verify that the generated dependency circle does not cause an infinite loop.
        Expected structure:

        id: 2 key: http://toolshed.local:10001__ESEP__filtering__ESEP__test__ESEP__871602b4276b
            ['http://toolshed.local:10001', 'emboss_5', 'test', '8de5fe0d7b04']
             id: 3 key: http://toolshed.local:10001__ESEP__emboss_datatypes__ESEP__test__ESEP__dbd4f68bf507
                 ['http://toolshed.local:10001', 'freebayes', 'test', 'f40028114098']
             id: 4 key: http://toolshed.local:10001__ESEP__freebayes__ESEP__test__ESEP__f40028114098
                 ['http://toolshed.local:10001', 'emboss_datatypes', 'test', 'dbd4f68bf507']
                 ['http://toolshed.local:10001', 'emboss_5', 'test', '8de5fe0d7b04']
                 ['http://toolshed.local:10001', 'column_maker', 'test', '83e956bdbac0']
             id: 5 key: http://toolshed.local:10001__ESEP__column_maker__ESEP__test__ESEP__83e956bdbac0
                 ['http://toolshed.local:10001', 'convert_chars', 'test', 'b28134220c8a']
             id: 6 key: http://toolshed.local:10001__ESEP__convert_chars__ESEP__test__ESEP__b28134220c8a
                 ['http://toolshed.local:10001', 'column_maker', 'test', '83e956bdbac0']
             id: 7 key: http://toolshed.local:10001__ESEP__emboss_5__ESEP__test__ESEP__8de5fe0d7b04
                 ['http://toolshed.local:10001', 'emboss_datatypes', 'test', 'dbd4f68bf507']
        """
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        filtering_repository = self._get_repository_by_name_and_owner(
            filtering_repository_name, common.test_user_1_name
        )
        freebayes_repository = self._get_repository_by_name_and_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
        convert_repository = self._get_repository_by_name_and_owner(convert_repository_name, common.test_user_1_name)
        bismark_repository = self._get_repository_by_name_and_owner(bismark_repository_name, common.test_user_1_name)
        self.check_repository_dependency(convert_repository, column_repository)
        self.check_repository_dependency(column_repository, convert_repository)
        self.check_repository_dependency(emboss_repository, bismark_repository)
        self.check_repository_dependency(filtering_repository, emboss_repository)
        for repository in [bismark_repository, emboss_repository, column_repository]:
            self.check_repository_dependency(freebayes_repository, repository)
        if not self.is_v2:
            strings_displayed = ["freebayes_0050 depends on freebayes_0050, emboss_0050, column_maker_0050."]
            self.display_manage_repository_page(freebayes_repository, strings_displayed=strings_displayed)

    @skip_if_api_v2
    def test_0050_verify_tool_dependencies(self):
        """Check that freebayes and emboss display tool dependencies."""
        freebayes_repository = self._get_repository_by_name_and_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        self.display_manage_repository_page(
            freebayes_repository,
            strings_displayed=["freebayes", "0.9.4_9696d0ce8a9", "samtools", "0.1.18", "Tool dependencies", "package"],
        )
        self.display_manage_repository_page(
            emboss_repository, strings_displayed=["Tool dependencies", "emboss", "5.0.0", "package"]
        )

    def test_0055_verify_repository_metadata(self):
        """Verify that resetting the metadata does not change it."""
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        filtering_repository = self._get_repository_by_name_and_owner(
            filtering_repository_name, common.test_user_1_name
        )
        freebayes_repository = self._get_repository_by_name_and_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        for repository in [emboss_repository, freebayes_repository, filtering_repository]:
            self.verify_unchanged_repository_metadata(repository)



import logging

import pytest

from ..base import common
from ..base.api import skip_if_api_v2
from ..base.twilltestcase import ShedTwillTestCase

repository_name = "filtering_0000"
repository_description = "Galaxy's filtering tool for test 0000"
repository_long_description = "Long description of Galaxy's filtering tool for test 0000"

log = logging.getLogger(__name__)


class TestBasicRepositoryFeatures(ShedTwillTestCase):
    """Test core repository features."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts and login as an admin user."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.test_user_2_email, username=common.test_user_2_name)
        self.login(email=common.admin_email, username=common.admin_username)

    @skip_if_api_v2
    # no replicating the functionality in tool shed 2.0, use Planemo
    # to create repositories.
    def test_0005_create_repository_without_categories(self):
        """Verify that a repository cannot be created unless at least one category has been defined."""
        strings_displayed = ["No categories have been configured in this instance of the Galaxy Tool Shed"]
        self.visit_url("/repository/create_repository")
        self.check_for_strings(strings_displayed=strings_displayed, strings_not_displayed=[])

    def test_0010_create_categories(self):
        """Create categories for this test suite"""
        self.create_category(
            name="Test 0000 Basic Repository Features 1", description="Test 0000 Basic Repository Features 1"
        )
        self.create_category(
            name="Test 0000 Basic Repository Features 2", description="Test 0000 Basic Repository Features 2"
        )

    def test_0015_create_repository(self):
        """Create the filtering repository"""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        category = self.populator.get_category_with_name("Test 0000 Basic Repository Features 1")
        strings_displayed = self.expect_repo_created_strings(repository_name)
        self.get_or_create_repository(
            name=repository_name,
            description=repository_description,
            long_description=repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=strings_displayed,
        )

    def test_0020_edit_repository(self):
        """Edit the repository name, description, and long description"""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        new_name = "renamed_filtering"
        new_description = "Edited filtering tool"
        new_long_description = "Edited long description"
        self.edit_repository_information(
            repository, repo_name=new_name, description=new_description, long_description=new_long_description
        )

    def test_0025_change_repository_category(self):
        """Change the categories associated with the filtering repository"""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.edit_repository_categories(
            repository,
            categories_to_add=["Test 0000 Basic Repository Features 2"],
            categories_to_remove=["Test 0000 Basic Repository Features 1"],
        )

    @skip_if_api_v2
    def test_0030_grant_write_access(self):
        """Grant write access to another user"""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.grant_write_access(repository, usernames=[common.test_user_2_name])
        self.revoke_write_access(repository, common.test_user_2_name)

    def test_0035_upload_filtering_1_1_0(self):
        """Upload filtering_1.1.0.tar to the repository"""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.commit_tar_to_repository(
            repository, "filtering/filtering_1.1.0.tar", commit_message="Uploaded filtering 1.1.0"
        )

    def test_0040_verify_repository(self):
        """Display basic repository pages"""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        latest_changeset_revision = self.get_repository_tip(repository)
        self.check_for_valid_tools(repository, strings_displayed=["Filter1"])
        self.check_count_of_metadata_revisions_associated_with_repository(repository, metadata_count=1)
        tip = self.get_repository_tip(repository)
        tool_guid = f"{self.url.replace('http://', '').rstrip('/')}/repos/user1/filtering_0000/Filter1/1.1.0"
        tool_metadata_strings_displayed = [
            tool_guid,
            "1.1.0",  # The tool version.
            "Filter1",  # The tool ID.
            "Filter",  # The tool name.
            "data on any column using simple expressions",
        ]  # The tool description.
        tool_page_strings_displayed = ["Filter (version 1.1.0)"]
        self.check_repository_tools_for_changeset_revision(
            repository,
            tip,
            tool_metadata_strings_displayed=tool_metadata_strings_displayed,
            tool_page_strings_displayed=tool_page_strings_displayed,
        )
        self.check_repository_metadata(repository, tip_only=False)
        self.browse_repository(
            repository, strings_displayed=[f"Repository '{repository.name}' revision", "(repository tip)"]
        )
        strings = ["Uploaded filtering 1.1.0"]
        if self._browser.is_twill:
            # this appears in a link - it isn't how one would check this
            # in playwright. But also we're testing the mercurial page
            # here so this is probably a questionable check overall.
            strings += [latest_changeset_revision]
        self.display_repository_clone_page(
            common.test_user_1_name,
            repository_name,
            strings_displayed=strings,
        )

    @skip_if_api_v2
    def test_0045_alter_repository_states(self):
        """Test toggling the malicious and deprecated repository flags."""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)
        self.set_repository_malicious(
            repository, set_malicious=True, strings_displayed=["The repository tip has been defined as malicious."]
        )
        self.set_repository_malicious(
            repository,
            set_malicious=False,
            strings_displayed=["The repository tip has been defined as <b>not</b> malicious."],
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.set_repository_deprecated(repository, strings_displayed=["has been marked as deprecated"])
        strings_displayed = ["This repository has been marked as deprecated", "Mark repository as not deprecated"]
        self.display_manage_repository_page(
            repository,
            strings_displayed=strings_displayed,
            strings_not_displayed=["Reset all repository metadata"],
        )
        self.browse_repository(repository)
        self.set_repository_deprecated(
            repository, strings_displayed=["has been marked as not deprecated"], set_deprecated=False
        )
        strings_displayed = ["Mark repository as deprecated", "Reset all repository metadata"]
        self.display_manage_repository_page(repository, strings_displayed=strings_displayed)

    @skip_if_api_v2
    # probably not porting this functionality - just test
    # with Twill for older UI and drop when that is all dropped
    def test_0050_display_repository_tip_file(self):
        """Display the contents of filtering.xml in the repository tip revision"""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        if self._browser.is_twill:
            self.display_repository_file_contents(
                repository=repository,
                filename="filtering.xml",
                filepath=None,
                strings_displayed=["1.1.0"],
                strings_not_displayed=[],
            )

    def test_0055_upload_filtering_txt_file(self):
        """Upload filtering.txt file associated with tool version 1.1.0."""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.add_file_to_repository(repository, "filtering/filtering_0000.txt")
        expected = self._escape_page_content_if_needed("Readme file for filtering 1.1.0")
        self.display_manage_repository_page(repository, strings_displayed=[expected])

    def test_0060_upload_filtering_test_data(self):
        """Upload filtering test data."""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.add_tar_to_repository(repository, "filtering/filtering_test_data.tar")
        if self._browser.is_twill:
            # probably not porting this functionality - just test
            # with Twill for older UI and drop when that is all dropped
            self.display_repository_file_contents(
                repository=repository,
                filename="1.bed",
                filepath="test-data",
                strings_displayed=[],
                strings_not_displayed=[],
            )
        self.check_repository_metadata(repository, tip_only=True)

    def test_0065_upload_filtering_2_2_0(self):
        """Upload filtering version 2.2.0"""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.add_tar_to_repository(repository, "filtering/filtering_2.2.0.tar")

    def test_0070_verify_filtering_repository(self):
        """Verify the new tool versions and repository metadata."""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        tip = self.get_repository_tip(repository)
        self.check_for_valid_tools(repository)
        if self.is_v2:
            strings_displayed = []
        else:
            strings_displayed = ["Select a revision"]
        self.display_manage_repository_page(repository, strings_displayed=strings_displayed)
        self.check_count_of_metadata_revisions_associated_with_repository(repository, metadata_count=2)
        tool_guid = f"{self.url.replace('http://', '').rstrip('/')}/repos/user1/filtering_0000/Filter1/2.2.0"
        tool_metadata_strings_displayed = [
            tool_guid,
            "2.2.0",  # The tool version.
            "Filter1",  # The tool ID.
            "Filter",  # The tool name.
            "data on any column using simple expressions",
        ]  # The tool description.
        tool_page_strings_displayed = ["Filter (version 2.2.0)"]
        self.check_repository_tools_for_changeset_revision(
            repository,
            tip,
            tool_metadata_strings_displayed=tool_metadata_strings_displayed,
            tool_page_strings_displayed=tool_page_strings_displayed,
        )
        self.check_repository_metadata(repository, tip_only=False)

    def test_0075_upload_readme_txt_file(self):
        """Upload readme.txt file associated with tool version 2.2.0."""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.add_file_to_repository(repository, "readme.txt")
        content = self._escape_page_content_if_needed("This is a readme file.")
        self.display_manage_repository_page(repository, strings_displayed=[content])
        # Verify that there is a different readme file for each metadata revision.
        readme_content = self._escape_page_content_if_needed("Readme file for filtering 1.1.0")
        self.display_manage_repository_page(
            repository,
            strings_displayed=[
                readme_content,
                content,
            ],
        )

    def test_0080_delete_readme_txt_file(self):
        """Delete the readme.txt file."""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.delete_files_from_repository(repository, filenames=["readme.txt"])
        self.check_count_of_metadata_revisions_associated_with_repository(repository, metadata_count=2)
        readme_content = self._escape_page_content_if_needed("Readme file for filtering 1.1.0")
        self.display_manage_repository_page(repository, strings_displayed=[readme_content])

    @skip_if_api_v2  # not re-implemented in the UI, there are API tests though
    def test_0085_search_for_valid_filter_tool(self):
        """Search for the filtering tool by tool ID, name, and version."""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        tip_changeset = self.get_repository_tip(repository)
        search_fields = dict(tool_id="Filter1", tool_name="filter", tool_version="2.2.0")
        self.search_for_valid_tools(
            search_fields=search_fields, strings_displayed=[tip_changeset], strings_not_displayed=[]
        )

    def test_0090_verify_repository_metadata(self):
        """Verify that resetting the metadata does not change it."""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.verify_unchanged_repository_metadata(repository)

    def test_0095_verify_reserved_repository_name_handling(self):
        """Check that reserved repository names are handled correctly."""
        category = self.populator.get_category_with_name("Test 0000 Basic Repository Features 1")
        error_message = (
            "The term 'repos' is a reserved word in the Tool Shed, so it cannot be used as a repository name."
        )
        with pytest.raises(AssertionError):
            self.get_or_create_repository(
                name="repos",
                description=repository_description,
                long_description=repository_long_description,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[error_message],
            )

    def test_0100_verify_reserved_username_handling(self):
        """Check that reserved usernames are handled correctly."""
        self.login(email="baduser@bx.psu.edu", username="repos")
        test_user_1 = self.test_db_util.get_user("baduser@bx.psu.edu")
        assert test_user_1 is None, 'Creating user with public name "repos" succeeded.'
        if not self.is_v2:
            # no longer use this terminology but the above test case ensures
            # the important thing and caught a bug in v2
            error_message = (
                "The term 'repos' is a reserved word in the Tool Shed, so it cannot be used as a public user name."
            )
            self.check_for_strings(strings_displayed=[error_message])

    def test_0105_contact_repository_owner(self):
        """"""
        # We no longer implement this.
        pass

    @skip_if_api_v2  # v2 doesn't implement repository deleting repositories
    def test_0110_delete_filtering_repository(self):
        """Delete the filtering_0000 repository and verify that it no longer has any downloadable revisions."""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)
        self.delete_repository(repository)
        metadata = self.populator.get_metadata(repository, downloadable_only=False)
        for _, value in metadata.root.items():
            assert not value.downloadable
        # Explicitly reload all metadata revisions from the database, to ensure that we have the current status of the downloadable flag.
        # for metadata_revision in repository.metadata_revisions:
        #    self.test_db_util.refresh(metadata_revision)
        # Marking a repository as deleted should result in no metadata revisions being downloadable.
        # assert True not in [metadata.downloadable for metadata in self._db_repository(repository).metadata_revisions]

    @skip_if_api_v2  # v2 doesn't implement repository deleting repositories
    def test_0115_undelete_filtering_repository(self):
        """Undelete the filtering_0000 repository and verify that it now has two downloadable revisions."""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)
        self.undelete_repository(repository)
        # Explicitly reload all metadata revisions from the database, to ensure that we have the current status of the downloadable flag.
        # for metadata_revision in repository.metadata_revisions:
        #    self.test_db_util.refresh(metadata_revision)
        # Marking a repository as undeleted should result in all previously downloadable metadata revisions being downloadable again.
        # In this case, there should be two downloadable revisions, one for filtering 1.1.0 and one for filtering 2.2.0.
        assert True in [metadata.downloadable for metadata in self._db_repository(repository).metadata_revisions]
        assert len(self._db_repository(repository).downloadable_revisions) == 2

    @skip_if_api_v2  # not re-implementing in tool shed 2.0
    def test_0120_enable_email_notifications(self):
        """Enable email notifications for test user 2 on filtering_0000."""
        # Log in as test_user_2
        self.login(email=common.test_user_2_email, username=common.test_user_2_name)
        # Get the repository, so we can pass the encoded repository id and browse_repositories method to the set_email_alerts method.
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        strings_displayed = ["Total alerts added: 1, total alerts removed: 0"]
        self.enable_email_alerts(repository, strings_displayed=strings_displayed)

    def test_0125_upload_new_readme_file(self):
        """Upload a new readme file to the filtering_0000 repository and verify that there is no error."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        # Upload readme.txt to the filtering_0000 repository and verify that it is now displayed.
        self.add_file_to_repository(repository, "filtering/readme.txt")
        content = self._escape_page_content_if_needed("These characters should not")
        self.display_manage_repository_page(repository, strings_displayed=[content])

    def test_0130_verify_handling_of_invalid_characters(self):
        """Load the above changeset in the change log and confirm that there is no server error displayed."""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        changeset_revision = self.get_repository_tip(repository)
        repository_id = repository.id
        changelog_tuples = self.get_repository_changelog_tuples(repository)
        revision_number = -1
        revision_hash = "000000000000"
        for numeric_changeset, changeset_hash in changelog_tuples:
            if str(changeset_hash) == str(changeset_revision):
                revision_number = numeric_changeset
                revision_hash = changeset_hash
                break
        # Check for the changeset revision, repository name, owner username, 'repos' in the clone url, and the captured
        # unicode decoding error message.
        content = self._escape_page_content_if_needed("These characters should not")
        strings_displayed = [
            f"{revision_number}:{revision_hash}",
            "filtering_0000",
            "user1",
            "repos",
            "added:",
            f"+{content}",
        ]
        self.load_changeset_in_tool_shed(repository_id, changeset_revision, strings_displayed=strings_displayed)

    def test_0135_api_get_repositories_in_category(self):
        """Load the api endpoint for repositories in a category."""
        categories = [
            self.populator.get_category_with_name(name)
            for name in ("Test 0000 Basic Repository Features 1", "Test 0000 Basic Repository Features 2")
        ]
        self.get_repositories_category_api(categories)

    def test_0140_view_invalid_changeset(self):
        """View repository using an invalid changeset"""
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        encoded_repository_id = repository.id
        assert encoded_repository_id
        strings_displayed = ["Invalid+changeset+revision"]
        view_repo_url = (
            f"/repository/view_repository?id={encoded_repository_id}&changeset_revision=nonsensical_changeset"
        )
        self.visit_url(view_repo_url)
        if self._browser.is_twill:
            self.check_for_strings(strings_displayed=strings_displayed, strings_not_displayed=[])



import logging

from galaxy.util.bunch import Bunch

log = logging.getLogger(__name__)


class RepositoryGridFilterManager:
    """Provides filtered views of the many Tool SHed repository grids."""

    filters = Bunch(
        CERTIFIED_LEVEL_ONE="certified_level_one",
        CERTIFIED_LEVEL_TWO="certified_level_two",
        CERTIFIED_LEVEL_ONE_SUITES="certified_level_one_suites",
        CERTIFIED_LEVEL_TWO_SUITES="certified_level_two_suites",
        SUITES="suites",
    )

    def get_grid_title(self, trans, trailing_string="", default=""):
        filter = self.get_filter(trans)
        if filter == self.filters.CERTIFIED_LEVEL_ONE:
            return f"Certified 1 Repositories {trailing_string}"
        if filter == self.filters.CERTIFIED_LEVEL_TWO:
            return f"Certified 2 Repositories {trailing_string}"
        if filter == self.filters.CERTIFIED_LEVEL_ONE_SUITES:
            return f"Certified 1 Repository Suites {trailing_string}"
        if filter == self.filters.CERTIFIED_LEVEL_TWO_SUITES:
            return f"Certified 2 Repository Suites {trailing_string}"
        if filter == self.filters.SUITES:
            return f"Repository Suites {trailing_string}"
        return f"{default} {trailing_string}"

    def get_filter(self, trans):
        filter = trans.get_cookie(name="toolshedrepogridfilter")
        return filter or None

    def is_valid_filter(self, filter):
        if filter is None:
            return True
        for valid_filter in self.filters.values():
            if filter == valid_filter:
                return True
        return False

    def set_filter(self, trans, **kwd):
        # Set a session cookie value with the selected filter.
        filter = kwd.get("filter", None)
        if filter is not None and self.is_valid_filter(filter):
            trans.set_cookie(value=filter, name="toolshedrepogridfilter")
        # if the filter is not valid, expire the cookie.
        trans.set_cookie(value=filter, name="toolshedrepogridfilter", age=-1)



import os

import pytest

from tool_shed.test.base import driver


@pytest.fixture(scope="session")
def embedded_driver():
    if not os.environ.get("GALAXY_TEST_ENVIRONMENT_CONFIGURED"):
        ts_driver = driver.ToolShedTestDriver()
        ts_driver.setup()
        try:
            yield ts_driver
        finally:
            ts_driver.tear_down()



from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

emboss_repository_name = "emboss_0050"
emboss_repository_description = "Galaxy's emboss tool"
emboss_repository_long_description = "Long description of Galaxy's emboss tool"

filtering_repository_name = "filtering_0050"
filtering_repository_description = "Galaxy's filtering tool"
filtering_repository_long_description = "Long description of Galaxy's filtering tool"

freebayes_repository_name = "freebayes_0050"
freebayes_repository_description = "Galaxy's freebayes tool"
freebayes_repository_long_description = "Long description of Galaxy's freebayes tool"

column_repository_name = "column_maker_0050"
column_repository_description = "Add column"
column_repository_long_description = "Compute an expression on every row"

convert_repository_name = "convert_chars_0050"
convert_repository_description = "Convert delimiters"
convert_repository_long_description = "Convert delimiters to tab"

bismark_repository_name = "bismark_0050"
bismark_repository_description = "A flexible aligner."
bismark_repository_long_description = "A flexible aligner and methylation caller for Bisulfite-Seq applications."

category_name = "Test 0050 Circular Dependencies 5 Levels"
category_description = "Test circular dependency features"

running_standalone = False


class TestInstallRepositoryCircularDependencies(ShedTwillTestCase):
    """Verify that the code correctly handles circular dependencies down to n levels."""

    requires_galaxy = True

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_convert_repository(self):
        """Create and populate convert_chars_0050."""
        category = self.create_category(name=category_name, description=category_description)
        global running_standalone
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=convert_repository_name,
            description=convert_repository_description,
            long_description=convert_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            running_standalone = True
            self.commit_tar_to_repository(
                repository,
                "convert_chars/convert_chars.tar",
                commit_message="Uploaded convert_chars tarball.",
            )

    def test_0010_create_column_repository(self):
        """Create and populate convert_chars_0050."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=column_repository_name,
            description=column_repository_description,
            long_description=column_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            self.commit_tar_to_repository(
                repository,
                "column_maker/column_maker.tar",
                commit_message="Uploaded column_maker tarball.",
            )

    def test_0015_create_emboss_datatypes_repository(self):
        """noop now..."""
        pass

    def test_0020_create_emboss_repository(self):
        """Create and populate emboss_0050."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=emboss_repository_name,
            description=emboss_repository_description,
            long_description=emboss_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            self.commit_tar_to_repository(
                repository,
                "emboss/emboss.tar",
                commit_message="Uploaded emboss tarball.",
            )

    def test_0025_create_filtering_repository(self):
        """Create and populate filtering_0050."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=filtering_repository_name,
            description=filtering_repository_description,
            long_description=filtering_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            self.commit_tar_to_repository(
                repository,
                "filtering/filtering_1.1.0.tar",
                commit_message="Uploaded filtering 1.1.0 tarball.",
            )

    def test_0030_create_freebayes_repository(self):
        """Create and populate freebayes_0050."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=freebayes_repository_name,
            description=freebayes_repository_description,
            long_description=freebayes_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            self.commit_tar_to_repository(
                repository,
                "freebayes/freebayes.tar",
                commit_message="Uploaded freebayes tarball.",
            )

    def test_0035_create_bismark_repository(self):
        """Create and populate bismark_0050."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=bismark_repository_name,
            description=bismark_repository_description,
            long_description=bismark_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            self.user_populator().setup_bismark_repo(repository, end=1)

    def test_0040_create_and_upload_dependency_definitions(self):
        """Set up the dependency structure."""
        if running_standalone:
            column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
            convert_repository = self._get_repository_by_name_and_owner(
                convert_repository_name, common.test_user_1_name
            )
            emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
            filtering_repository = self._get_repository_by_name_and_owner(
                filtering_repository_name, common.test_user_1_name
            )
            freebayes_repository = self._get_repository_by_name_and_owner(
                freebayes_repository_name, common.test_user_1_name
            )
            bismark_repository = self._get_repository_by_name_and_owner(
                bismark_repository_name, common.test_user_1_name
            )
            dependency_xml_path = self.generate_temp_path("test_1050", additional_paths=["dependencies"])
            # convert_chars depends on column_maker
            # column_maker depends on convert_chars
            # emboss depends on bismark
            # freebayes depends on freebayes, emboss, bismark, and column_maker
            # filtering depends on emboss
            column_tuple = (
                self.url,
                column_repository.name,
                column_repository.owner,
                self.get_repository_tip(column_repository),
            )
            convert_tuple = (
                self.url,
                convert_repository.name,
                convert_repository.owner,
                self.get_repository_tip(convert_repository),
            )
            freebayes_tuple = (
                self.url,
                freebayes_repository.name,
                freebayes_repository.owner,
                self.get_repository_tip(freebayes_repository),
            )
            emboss_tuple = (
                self.url,
                emboss_repository.name,
                emboss_repository.owner,
                self.get_repository_tip(emboss_repository),
            )
            bismark_tuple = (
                self.url,
                bismark_repository.name,
                bismark_repository.owner,
                self.get_repository_tip(bismark_repository),
            )
            self.create_repository_dependency(
                repository=convert_repository, repository_tuples=[column_tuple], filepath=dependency_xml_path
            )
            self.create_repository_dependency(
                repository=column_repository, repository_tuples=[convert_tuple], filepath=dependency_xml_path
            )
            self.create_repository_dependency(
                repository=emboss_repository, repository_tuples=[bismark_tuple], filepath=dependency_xml_path
            )
            self.create_repository_dependency(
                repository=freebayes_repository,
                repository_tuples=[freebayes_tuple, emboss_tuple, column_tuple],
                filepath=dependency_xml_path,
            )
            self.create_repository_dependency(
                repository=filtering_repository, repository_tuples=[emboss_tuple], filepath=dependency_xml_path
            )

    def test_0045_verify_repository_dependencies(self):
        """Verify that the generated dependency circle does not cause an infinite loop.
        Expected structure:

        id: 2 key: http://toolshed.local:10001__ESEP__filtering__ESEP__test__ESEP__871602b4276b
            ['http://toolshed.local:10001', 'emboss_5', 'test', '8de5fe0d7b04']
             id: 3 key: http://toolshed.local:10001__ESEP__emboss_datatypes__ESEP__test__ESEP__dbd4f68bf507
                 ['http://toolshed.local:10001', 'freebayes', 'test', 'f40028114098']
             id: 4 key: http://toolshed.local:10001__ESEP__freebayes__ESEP__test__ESEP__f40028114098
                 ['http://toolshed.local:10001', 'emboss_datatypes', 'test', 'dbd4f68bf507']
                 ['http://toolshed.local:10001', 'emboss_5', 'test', '8de5fe0d7b04']
                 ['http://toolshed.local:10001', 'column_maker', 'test', '83e956bdbac0']
             id: 5 key: http://toolshed.local:10001__ESEP__column_maker__ESEP__test__ESEP__83e956bdbac0
                 ['http://toolshed.local:10001', 'convert_chars', 'test', 'b28134220c8a']
             id: 6 key: http://toolshed.local:10001__ESEP__convert_chars__ESEP__test__ESEP__b28134220c8a
                 ['http://toolshed.local:10001', 'column_maker', 'test', '83e956bdbac0']
             id: 7 key: http://toolshed.local:10001__ESEP__emboss_5__ESEP__test__ESEP__8de5fe0d7b04
                 ['http://toolshed.local:10001', 'emboss_datatypes', 'test', 'dbd4f68bf507']
        """
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        filtering_repository = self._get_repository_by_name_and_owner(
            filtering_repository_name, common.test_user_1_name
        )
        freebayes_repository = self._get_repository_by_name_and_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        column_repository = self._get_repository_by_name_and_owner(column_repository_name, common.test_user_1_name)
        convert_repository = self._get_repository_by_name_and_owner(convert_repository_name, common.test_user_1_name)
        bismark_repository = self._get_repository_by_name_and_owner(bismark_repository_name, common.test_user_1_name)
        self.check_repository_dependency(convert_repository, column_repository)
        self.check_repository_dependency(column_repository, convert_repository)
        self.check_repository_dependency(emboss_repository, bismark_repository)
        self.check_repository_dependency(filtering_repository, emboss_repository)
        for repository in [bismark_repository, emboss_repository, column_repository]:
            self.check_repository_dependency(freebayes_repository, repository)
        freebayes_dependencies = [
            freebayes_repository,
            emboss_repository,
            column_repository,
        ]
        strings_displayed = [
            f"{freebayes_repository.name} depends on {', '.join(repo.name for repo in freebayes_dependencies)}."
        ]
        if not self.is_v2:
            self.display_manage_repository_page(freebayes_repository, strings_displayed=strings_displayed)

    def test_0050_verify_tool_dependencies(self):
        """Check that freebayes and emboss display tool dependencies."""
        freebayes_repository = self._get_repository_by_name_and_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        if not self.is_v2:
            self.display_manage_repository_page(
                freebayes_repository,
                strings_displayed=["freebayes", "0.9.4_9696d0ce8a9", "samtools", "0.1.18", "Tool dependencies"],
            )
            self.display_manage_repository_page(
                emboss_repository, strings_displayed=["Tool dependencies", "emboss", "5.0.0", "package"]
            )

    def test_0055_install_column_repository(self):
        """Install column_maker with repository dependencies."""
        self._install_repository(
            column_repository_name,
            common.test_user_1_name,
            category_name,
            install_tool_dependencies=False,
            install_repository_dependencies=True,
            new_tool_panel_section_label="column_maker",
        )
        # This should result in column_maker and convert_chars being installed, and the rest never installed.
        installed_repositories = [
            (column_repository_name, common.test_user_1_name),
            (convert_repository_name, common.test_user_1_name),
        ]
        strings_displayed = ["column_maker_0050", "convert_chars_0050"]
        strings_not_displayed = [
            "emboss_0050",
            "filtering_0050",
            "freebayes_0050",
            "bismark_0050",
        ]
        self._assert_has_installed_repos_with_names(*strings_displayed)
        self._assert_has_no_installed_repos_with_names(*strings_not_displayed)
        self.verify_installed_repositories(installed_repositories=installed_repositories)

    def test_0060_install_emboss_repository(self):
        """Install emboss_5 with repository dependencies."""
        self._install_repository(
            emboss_repository_name,
            common.test_user_1_name,
            category_name,
            install_tool_dependencies=False,
            install_repository_dependencies=True,
            new_tool_panel_section_label="emboss_5_0050",
        )
        # Now we have emboss, bismark, column_maker, and convert_chars installed, filtering and freebayes never installed.
        installed_repositories = [
            (column_repository_name, common.test_user_1_name),
            (emboss_repository_name, common.test_user_1_name),
            (convert_repository_name, common.test_user_1_name),
            (bismark_repository_name, common.test_user_1_name),
        ]
        strings_displayed = [
            "emboss_0050",
            "column_maker_0050",
            "convert_chars_0050",
            "bismark_0050",
        ]
        strings_not_displayed = ["filtering_0050", "freebayes_0050"]
        self._assert_has_installed_repos_with_names(*strings_displayed)
        self._assert_has_no_installed_repos_with_names(*strings_not_displayed)
        self.verify_installed_repositories(installed_repositories)

    def test_0065_deactivate_bismark_repository(self):
        """Deactivate bismark and verify things are okay."""
        repository = self._get_installed_repository_by_name_owner(bismark_repository_name, common.test_user_1_name)
        self.deactivate_repository(repository)
        # Now we have emboss, bismark, column_maker, and convert_chars installed, filtering and freebayes never installed.
        installed_repositories = [
            (column_repository_name, common.test_user_1_name),
            (emboss_repository_name, common.test_user_1_name),
            (convert_repository_name, common.test_user_1_name),
        ]
        strings_displayed = ["emboss_0050", "column_maker_0050", "convert_chars_0050"]
        strings_not_displayed = ["bismark", "filtering_0050", "freebayes_0050"]
        self._assert_has_installed_repos_with_names(*strings_displayed)
        self._assert_has_no_installed_repos_with_names(*strings_not_displayed)
        self.verify_installed_repositories(installed_repositories)

    def test_0070_uninstall_emboss_repository(self):
        """Uninstall the emboss_5 repository."""
        repository = self._get_installed_repository_by_name_owner(emboss_repository_name, common.test_user_1_name)
        self._uninstall_repository(repository)
        self._assert_has_no_installed_repos_with_names(repository.name)
        self._refresh_tool_shed_repository(repository)
        self.check_galaxy_repository_tool_panel_section(repository, "emboss_5_0050")
        # Now we have bismark, column_maker, and convert_chars installed, filtering and freebayes never installed,
        # and emboss uninstalled.
        installed_repositories = [
            (column_repository_name, common.test_user_1_name),
            (convert_repository_name, common.test_user_1_name),
        ]
        strings_displayed = ["column_maker_0050", "convert_chars_0050"]
        strings_not_displayed = ["emboss_0050", "filtering_0050", "freebayes_0050"]
        self._assert_has_installed_repos_with_names(*strings_displayed)
        self._assert_has_no_installed_repos_with_names(*strings_not_displayed)
        self.verify_installed_repositories(installed_repositories)

    def test_0075_install_freebayes_repository(self):
        """Install freebayes with repository dependencies. This should also automatically reactivate bismark and reinstall emboss_5."""
        strings_displayed = ["Handle", "tool dependencies", "freebayes", "0.9.4_9696d0ce8a9", "samtools", "0.1.18"]
        self._install_repository(
            freebayes_repository_name,
            common.test_user_1_name,
            category_name,
            install_tool_dependencies=False,
            install_repository_dependencies=True,
            new_tool_panel_section_label="freebayes",
        )
        self._assert_has_installed_repos_with_names("emboss_0050")
        # Installing freebayes should automatically reinstall emboss and reactivate bismark.
        # Now column_maker, convert_chars, emboss, freebayes, and bismark should be installed.
        installed_repositories = [
            (column_repository_name, common.test_user_1_name),
            (emboss_repository_name, common.test_user_1_name),
            (freebayes_repository_name, common.test_user_1_name),
            (convert_repository_name, common.test_user_1_name),
            (bismark_repository_name, common.test_user_1_name),
        ]
        strings_displayed = [
            "emboss_0050",
            "column_maker_0050",
            "convert_chars_0050",
            "bismark_0050",
            "freebayes_0050",
        ]
        strings_not_displayed = ["filtering_0050"]
        self._assert_has_installed_repos_with_names(*strings_displayed)
        self._assert_has_no_installed_repos_with_names(*strings_not_displayed)
        self.verify_installed_repositories(installed_repositories)



import logging

from ..base import common
from ..base.api import skip_if_api_v2
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

repository_name = "filtering_0530"
repository_description = "Filtering repository for test 0530"
repository_long_description = "This is the filtering repository for test 0530."
category_name = "Test 0530 Repository Admin Role"
category_description = "Verify the functionality of the code that handles the repository admin role."

"""
1. Create new repository as user user1.

2. Check to make sure a new role was created named <repo_name>_user1_admin.

3. Check to make sure a new repository_role_association record was created with appropriate repository id and role id.

4. Change the name of the repository created in step 1 - this can be done as long as the repository has not been installed or cloned.

5. Make sure the name of the role initially inspected in Step 2 has been changed to reflect the new repository name from Step 4.

6. Log into the Tool Shed as a user that is not the repository owner (e.g., user2) and make sure the repository name
   and description cannot be changed.

7. As user user1, add user user2 as a repository admin user.

8. Log into the Tool Shed as user user2 and make sure the repository name and description can now be changed.
"""


class TestRepositoryAdminRole(ShedTwillTestCase):
    """Verify that the code correctly handles the repository admin role."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.test_user_2_email, username=common.test_user_2_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_filtering_repository(self):
        """Create and populate the filtering_0530 repository.

        This is step 1 - Create new repository as user user1.
        """
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=repository_name,
            description=repository_description,
            long_description=repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "filtering/filtering_1.1.0.tar",
            commit_message="Uploaded filtering 1.1.0 tarball.",
        )

    def test_0010_verify_repository_admin_role_exists(self):
        """Verify that the role filtering_0530_user1_admin exists.

        This is step 2 - Check to make sure a new role was created named filtering_0530_user1_admin.
        """
        test_user_1 = self.test_db_util.get_user(common.test_user_1_email)
        repository_admin_role = self.test_db_util.get_role(test_user_1, "filtering_0530_user1_admin")
        assert repository_admin_role is not None, "Admin role for filtering_0530 was not created."

    def test_0015_verify_repository_role_association(self):
        """Verify that the filtering_0530_user1_admin role is associated with the filtering_0530 repository.

        This is step 3 - Check to make sure a new repository_role_association record was created with appropriate repository id and role id.
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        test_user_1 = self.test_db_util.get_user(common.test_user_1_email)
        repository_admin_role = self.test_db_util.get_role(test_user_1, "filtering_0530_user1_admin")
        repository_role_association = self.test_db_util.get_repository_role_association(
            self._db_repository(repository).id, repository_admin_role.id
        )
        assert (
            repository_role_association is not None
        ), "Repository filtering_0530 is not associated with the filtering_0530_user1_admin role."

    def test_0020_rename_repository(self):
        """Rename the repository to renamed_filtering_0530.

        This is step 4 - Change the name of the repository created in step 1 - this can be done as long as the repository has not
        been installed or cloned.
        """
        repository = self._get_repository_by_name_and_owner(repository_name, common.test_user_1_name)
        self.edit_repository_information(repository, revert=False, repo_name="renamed_filtering_0530")
        repository = self._get_repository_by_name_and_owner("renamed_filtering_0530", common.test_user_1_name)
        assert repository.name == "renamed_filtering_0530", "Repository was not renamed to renamed_filtering_0530."

    @skip_if_api_v2
    def test_0030_verify_access_denied(self):
        """Make sure a non-admin user can't modify the repository.

        This is step 6 - Log into the Tool Shed as a user that is not the repository owner (e.g., user2) and make sure the repository
        name and description cannot be changed.
        """
        self.login(email=common.test_user_2_email, username=common.test_user_2_name)
        repository = self._get_repository_by_name_and_owner("renamed_filtering_0530", common.test_user_1_name)
        strings_not_displayed = ["Manage repository"]
        strings_displayed = ["View repository"]
        self.display_manage_repository_page(repository, strings_not_displayed=strings_not_displayed)
        self.submit_form(form_no=0, button="edit_repository_button", description="This description has been modified.")
        strings_displayed = ["You are not the owner of this repository, so you cannot administer it."]
        strings_not_displayed = ["The repository information has been updated."]
        self.check_for_strings(strings_displayed=strings_displayed, strings_not_displayed=strings_not_displayed)

    def test_0035_grant_admin_role(self):
        """Grant the repository admin role to user2.

        This is step 7 - As user user1, add user user2 as a repository admin user.
        """
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        test_user_2 = self.test_db_util.get_user(common.test_user_2_email)
        repository = self._get_repository_by_name_and_owner("renamed_filtering_0530", common.test_user_1_name)
        self.assign_admin_role(repository, test_user_2)

    def test_0040_rename_repository_as_repository_admin(self):
        """Rename the repository as user2.

        This is step 8 - Log into the Tool Shed as user user2 and make sure the repository name and description can now be changed.
        """
        self.login(email=common.test_user_2_email, username=common.test_user_2_name)
        repository = self._get_repository_by_name_and_owner("renamed_filtering_0530", common.test_user_1_name)
        self.edit_repository_information(repository, revert=False, repo_name="filtering_0530")
        repository = self._get_repository_by_name_and_owner("filtering_0530", common.test_user_1_name)
        assert repository.name == "filtering_0530", "User with admin role failed to rename repository."
        test_user_1 = self.test_db_util.get_user(common.test_user_1_email)
        old_repository_admin_role = self.test_db_util.get_role(test_user_1, "renamed_filtering_0530_user1_admin")
        assert old_repository_admin_role is None, "Admin role renamed_filtering_0530_user1_admin incorrectly exists."
        new_repository_admin_role = self.test_db_util.get_role(test_user_1, "filtering_0530_user1_admin")
        assert new_repository_admin_role is not None, "Admin role filtering_0530_user1_admin does not exist."



import logging

from galaxy.tool_shed.repository_type import (
    REPOSITORY_DEPENDENCY_DEFINITION_FILENAME,
    REPOSITORY_SUITE_DEFINITION,
    TOOL_DEPENDENCY_DEFINITION,
    TOOL_DEPENDENCY_DEFINITION_FILENAME,
    types,
    UNRESTRICTED,
)
from galaxy.web.form_builder import SelectField

log = logging.getLogger(__name__)


def build_repository_type_select_field(trans, repository=None, name="repository_type"):
    """Called from the Tool Shed to generate the current list of supported repository types."""
    if repository:
        selected_type = str(repository.type)
    else:
        selected_type = None
    repository_type_select_field = SelectField(name=name)
    for type_class in trans.app.repository_types_registry.repository_types_by_label.values():
        option_label = str(type_class.label)
        option_value = str(type_class.type)
        if selected_type and selected_type == option_value:
            selected = True
        else:
            selected = False
        if repository:
            if repository.type == option_value:
                repository_type_select_field.add_option(option_label, option_value, selected=selected)
            elif type_class.is_valid_for_type(repository):
                repository_type_select_field.add_option(option_label, option_value, selected=selected)
        else:
            repository_type_select_field.add_option(option_label, option_value, selected=selected)
    return repository_type_select_field


def generate_message_for_repository_type_change(app, repository):
    message = ""
    if repository.can_change_type_to(app, REPOSITORY_SUITE_DEFINITION):
        repository_suite_definition_type_class = app.repository_types_registry.get_class_by_label(
            REPOSITORY_SUITE_DEFINITION
        )
        message = (
            f"This repository currently contains a single file named <b>{REPOSITORY_DEPENDENCY_DEFINITION_FILENAME}</b>.  If the intent of this repository is "
            "to define relationships to a collection of repositories that contain related Galaxy utilities with "
            f"no plans to add additional files, consider setting its type to <b>{repository_suite_definition_type_class.label}</b>.<br/>"
        )
    elif repository.can_change_type_to(app, TOOL_DEPENDENCY_DEFINITION):
        tool_dependency_definition_type_class = app.repository_types_registry.get_class_by_label(
            TOOL_DEPENDENCY_DEFINITION
        )
        message = (
            f"This repository currently contains a single file named <b>{TOOL_DEPENDENCY_DEFINITION_FILENAME}</b>.  If additional files will "
            f"not be added to this repository, consider setting its type to <b>{tool_dependency_definition_type_class.label}</b>.<br/>"
        )
    return message


__all__ = (
    "build_repository_type_select_field",
    "generate_message_for_repository_type_change",
    "REPOSITORY_DEPENDENCY_DEFINITION_FILENAME",
    "REPOSITORY_SUITE_DEFINITION",
    "TOOL_DEPENDENCY_DEFINITION",
    "TOOL_DEPENDENCY_DEFINITION_FILENAME",
    "UNRESTRICTED",
    "types",
)



from typing import (
    Optional,
    TYPE_CHECKING,
)

from routes import url_for

from galaxy.util.tool_shed.common_util import (
    accumulate_tool_dependencies,
    check_tool_tag_set,
    generate_clone_url_for_installed_repository,
    generate_clone_url_from_repo_info_tup,
    get_protocol_from_tool_shed_url,
    get_repository_dependencies,
    get_tool_shed_repository_ids,
    get_tool_shed_repository_url,
    get_tool_shed_url_from_tool_shed_registry,
    handle_galaxy_url,
    handle_tool_shed_url_protocol,
    parse_repository_dependency_tuple,
    remove_port_from_tool_shed_url,
    remove_protocol_and_port_from_tool_shed_url,
    remove_protocol_and_user_from_clone_url,
    remove_protocol_from_tool_shed_url,
)

if TYPE_CHECKING:
    from tool_shed.context import ProvidesRepositoriesContext
    from tool_shed.webapp.model import (
        Repository,
        User,
    )


def generate_clone_url_for(trans: "ProvidesRepositoriesContext", repository: "Repository") -> str:
    return generate_clone_url_for_repository_in_tool_shed(trans.user, repository, trans.repositories_hostname)


def generate_clone_url_for_repository_in_tool_shed(
    user: Optional["User"], repository: "Repository", hostname: Optional[str] = None
) -> str:
    """Generate the URL for cloning a repository that is in the tool shed."""
    base_url = hostname or url_for("/", qualified=True).rstrip("/")
    return f"{base_url}/repos/{repository.user.username}/{repository.name}"


__all__ = (
    "accumulate_tool_dependencies",
    "check_tool_tag_set",
    "generate_clone_url_for_installed_repository",
    "generate_clone_url_for_repository_in_tool_shed",
    "generate_clone_url_from_repo_info_tup",
    "get_repository_dependencies",
    "get_protocol_from_tool_shed_url",
    "get_tool_shed_repository_ids",
    "get_tool_shed_url_from_tool_shed_registry",
    "get_tool_shed_repository_url",
    "handle_galaxy_url",
    "handle_tool_shed_url_protocol",
    "parse_repository_dependency_tuple",
    "remove_port_from_tool_shed_url",
    "remove_protocol_and_port_from_tool_shed_url",
    "remove_protocol_and_user_from_clone_url",
    "remove_protocol_from_tool_shed_url",
)



import os
from functools import wraps
from typing import (
    Any,
    Dict,
    Optional,
)

import pytest

from galaxy.tool_util.verify.interactor import GalaxyInteractorApi
from galaxy_test.base.api_util import (
    get_admin_api_key as get_galaxy_admin_api_key,
    get_user_api_key as get_galaxy_user_key,
    TEST_USER,
)
from galaxy_test.base.uses_shed_api import UsesShedApi
from galaxy_test.driver.testcase import DrivenFunctionalTestCase
from . import driver
from .api_util import (
    ensure_user_with_email,
    get_admin_api_key,
    get_user_api_key,
    ShedApiInteractor,
)
from .populators import ToolShedPopulator


class ShedBaseTestCase(DrivenFunctionalTestCase):
    _populator: Optional[ToolShedPopulator] = None

    @property
    def populator(self) -> ToolShedPopulator:
        if self._populator is None:
            self._populator = self._get_populator(self.api_interactor)
        return self._populator

    @property
    def admin_api_interactor(self) -> ShedApiInteractor:
        return ShedApiInteractor(self.url, get_admin_api_key())

    def _api_interactor_for_key(self, key: str) -> ShedApiInteractor:
        return self._api_interactor(key)

    def populator_for_key(self, key: str) -> ToolShedPopulator:
        return self._get_populator(self._api_interactor_for_key(key))

    @property
    def api_interactor(self) -> ShedApiInteractor:
        user_api_key = get_user_api_key()
        if user_api_key is None:
            email = TEST_USER
            password = "testpassword"
            ensure_user_with_email(self.admin_api_interactor, email, password)
            user_api_key = self.admin_api_interactor.create_api_key(email, password)
        return self._api_interactor_for_key(user_api_key)

    def _api_interactor_by_credentials(self, email: str, password: str) -> ShedApiInteractor:
        ensure_user_with_email(self.admin_api_interactor, email, password)
        user_api_key = self.admin_api_interactor.create_api_key(email, password)
        return self._api_interactor(user_api_key)

    def _api_interactor(self, api_key: str) -> ShedApiInteractor:
        return ShedApiInteractor(self.url, api_key)

    def _get_populator(self, user_api_interactor) -> ToolShedPopulator:
        return ToolShedPopulator(self.admin_api_interactor, user_api_interactor)

    def setUp(self):
        host = os.environ.get("TOOL_SHED_TEST_HOST")
        assert host
        self.host = host
        self.port = os.environ.get("TOOL_SHED_TEST_PORT")
        self.url = f"http://{self.host}:{self.port}"
        self.galaxy_host = os.environ.get("GALAXY_TEST_HOST")
        self.galaxy_port = os.environ.get("GALAXY_TEST_PORT")
        self.galaxy_url = f"http://{self.galaxy_host}:{self.galaxy_port}"

    __test__ = True
    galaxy_driver_class = driver.ToolShedTestDriver

    @classmethod
    def setUpClass(cls):
        """Configure and start Galaxy for a test."""
        pass

    @classmethod
    def tearDownClass(cls):
        """Shutdown Galaxy server and cleanup temp directory."""
        pass

    @pytest.fixture(autouse=True)
    def _get_driver(self, embedded_driver):
        self._test_driver = embedded_driver


class ShedGalaxyInteractorApi(GalaxyInteractorApi):
    def __init__(self, galaxy_url: str):
        interactor_kwds: Dict[str, Any] = {}
        interactor_kwds["galaxy_url"] = galaxy_url
        interactor_kwds["master_api_key"] = get_galaxy_admin_api_key()
        interactor_kwds["api_key"] = get_galaxy_user_key()
        super().__init__(**interactor_kwds)


def make_skip_if_api_version_wrapper(version):
    def wrapper(method):
        @wraps(method)
        def wrapped_method(api_test_case, *args, **kwd):
            interactor: ShedApiInteractor = api_test_case.api_interactor
            api_version = interactor.api_version
            if api_version == version:
                raise pytest.skip(f"{version} tool shed API found, skipping test")
            return method(api_test_case, *args, **kwd)

        return wrapped_method

    return wrapper


skip_if_api_v1 = make_skip_if_api_version_wrapper("v1")
skip_if_api_v2 = make_skip_if_api_version_wrapper("v2")


class ShedApiTestCase(ShedBaseTestCase, UsesShedApi):
    _galaxy_interactor: Optional[GalaxyInteractorApi] = None

    @property
    def galaxy_interactor(self) -> GalaxyInteractorApi:
        if self._galaxy_interactor is None:
            self._galaxy_interactor = ShedGalaxyInteractorApi(self.galaxy_url)
        return self._galaxy_interactor



from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

repo_name = "filtering_0000"
repo_description = "Galaxy's filtering tool"


class TestBasicToolShedFeatures(ShedTwillTestCase):
    """Test installing a basic repository."""

    requires_galaxy = True

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_ensure_repositories_and_categories_exist(self):
        """Create the 0000 category and upload the filtering repository to it, if necessary."""
        self.login(email=common.admin_email, username=common.admin_username)
        category = self.create_category(
            name="Test 0000 Basic Repository Features 2",
            description="Test Description 0000 Basic Repository Features 2",
        )
        category = self.create_category(
            name="Test 0000 Basic Repository Features 1",
            description="Test Description 0000 Basic Repository Features 1",
        )
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=repo_name,
            description=repo_description,
            long_description="Long description of Galaxy's filtering tool",
            owner=common.test_user_1_name,
            category=category,
        )
        if self.repository_is_new(repository):
            self.commit_tar_to_repository(
                repository,
                "filtering/filtering_1.1.0.tar",
                commit_message="Uploaded filtering 1.1.0 tarball.",
            )
            self.add_file_to_repository(repository, "filtering/filtering_0000.txt")
            self.add_tar_to_repository(repository, "filtering/filtering_2.2.0.tar")
            self.add_file_to_repository(repository, "readme.txt")

    def test_0010_browse_tool_sheds(self):
        """Browse the available tool sheds in this Galaxy instance."""
        self.browse_tool_shed(
            url=self.url,
            strings_displayed=["Test 0000 Basic Repository Features 1", "Test 0000 Basic Repository Features 2"],
        )

    def test_0015_browse_test_0000_category(self):
        """Browse the category created in test 0000. It should contain the filtering_0000 repository also created in that test."""
        category = self.populator.get_category_with_name("Test 0000 Basic Repository Features 1")
        self.browse_category(category, strings_displayed=[repo_name])

    def test_0020_preview_filtering_repository(self):
        """Load the preview page for the filtering_0000 repository in the tool shed."""
        self.preview_repository_in_tool_shed(
            repo_name, common.test_user_1_name, strings_displayed=[repo_name, "Valid tools"]
        )

    def test_0025_install_filtering_repository(self):
        self._install_repository(
            repo_name,
            common.test_user_1_name,
            "Test 0000 Basic Repository Features 1",
            new_tool_panel_section_label="test_1000",
        )
        installed_repository = self._get_installed_repository_by_name_owner(repo_name, common.test_user_1_name)
        changeset = str(installed_repository.installed_changeset_revision)
        assert self._get_installed_repository_for(common.test_user_1, repo_name, changeset)
        self._assert_has_valid_tool_with_name("Filter1")
        self._assert_repo_has_tool_with_id(installed_repository, "Filter1")

    def test_0030_install_filtering_repository_again(self):
        """Attempt to install the already installed filtering repository."""
        installed_repository = self._get_installed_repository_by_name_owner(repo_name, common.test_user_1_name)
        # Just make sure the repo is still installed, used to monitoring tests but we've
        # removed that page.
        self._install_repository(
            repo_name,
            common.test_user_1_name,
            "Test 0000 Basic Repository Features 1",
        )
        changeset = str(installed_repository.installed_changeset_revision)
        assert self._get_installed_repository_for(common.test_user_1, repo_name, changeset)

    def test_0035_verify_installed_repository_metadata(self):
        """Verify that resetting the metadata on an installed repository does not change the metadata."""
        self.verify_installed_repository_metadata_unchanged(repo_name, common.test_user_1_name)



from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

datatypes_repository_name = "blast_datatypes_0120"
datatypes_repository_description = "Galaxy applicable datatypes for BLAST"
datatypes_repository_long_description = "Galaxy datatypes for the BLAST top hit descriptons tool"

tool_repository_name = "blastxml_to_top_descr_0120"
tool_repository_description = "BLAST top hit descriptions"
tool_repository_long_description = "Make a table from BLAST XML"

"""
Tool shed side:

1) Create and populate blast_datatypes_0120.
1a) Check for appropriate strings.
2) Create and populate blastxml_to_top_descr_0120.
2a) Check for appropriate strings.
3) Upload repository_dependencies.xml to blastxml_to_top_descr_0120 that defines a relationship to blast_datatypes_0120.
3a) Check for appropriate strings.
"""


class TestRepositoryMultipleOwners(ShedTwillTestCase):
    def test_0000_initiate_users(self):
        """Create necessary user accounts and login as an admin user.

        Create all the user accounts that are needed for this test script to run independently of other tests.
        Previously created accounts will not be re-created.
        """
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.test_user_2_email, username=common.test_user_2_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_datatypes_repository(self):
        """Create and populate the blast_datatypes_0120 repository

        We are at step 1.
        Create and populate blast_datatypes.
        """
        category = self.create_category(name="Test 0120", description="Description of test 0120")
        self.login(email=common.test_user_2_email, username=common.test_user_2_name)
        strings_displayed = self.expect_repo_created_strings(datatypes_repository_name)
        repository = self.get_or_create_repository(
            name=datatypes_repository_name,
            description=datatypes_repository_description,
            long_description=datatypes_repository_long_description,
            owner=common.test_user_2_name,
            category=category,
            strings_displayed=strings_displayed,
        )
        self.commit_tar_to_repository(
            repository,
            "blast/blast_datatypes.tar",
            commit_message="Uploaded blast_datatypes tarball.",
        )

    def test_0010_verify_datatypes_repository(self):
        """Verify the blast_datatypes_0120 repository.

        We are at step 1a.
        Check for appropriate strings, most importantly BlastXml, BlastNucDb, and BlastProtDb,
        the datatypes that are defined in datatypes_conf.xml.
        """
        repository = self._get_repository_by_name_and_owner(datatypes_repository_name, common.test_user_2_name)
        # v2 rightfully doesn't display anything about datatypes...
        strings_displayed = ["Galaxy datatypes for the BLAST top hit"]
        self.display_manage_repository_page(repository, strings_displayed=strings_displayed)

    def test_0015_create_tool_repository(self):
        """Create and populate the blastxml_to_top_descr_0120 repository

        We are at step 2.
        Create and populate blastxml_to_top_descr_0120.
        """
        category = self.create_category(name="Test 0120", description="Description of test 0120")
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        strings_displayed = self.expect_repo_created_strings(tool_repository_name)
        repository = self.get_or_create_repository(
            name=tool_repository_name,
            description=tool_repository_description,
            long_description=tool_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=strings_displayed,
        )
        self.commit_tar_to_repository(
            repository,
            "blast/blastxml_to_top_descr.tar",
            commit_message="Uploaded blastxml_to_top_descr tarball.",
        )

    def test_0020_verify_tool_repository(self):
        """Verify the blastxml_to_top_descr_0120 repository.

        We are at step 2a.
        Check for appropriate strings, such as tool name, description, and version.
        """
        repository = self._get_repository_by_name_and_owner(tool_repository_name, common.test_user_1_name)
        strings_displayed = ["blastxml_to_top_descr_0120", "BLAST top hit descriptions", "Make a table from BLAST XML"]
        strings_displayed.append("0.0.1")
        if not self.is_v2:
            strings_displayed.append("Valid tools")
        self.display_manage_repository_page(repository, strings_displayed=strings_displayed)

    def test_0025_create_repository_dependency(self):
        """Create a repository dependency on blast_datatypes_0120.

        We are at step 3.
        Create a simple repository dependency for blastxml_to_top_descr_0120 that defines a dependency on blast_datatypes_0120.
        """
        datatypes_repository = self._get_repository_by_name_and_owner(
            datatypes_repository_name, common.test_user_2_name
        )
        tool_repository = self._get_repository_by_name_and_owner(tool_repository_name, common.test_user_1_name)
        dependency_xml_path = self.generate_temp_path("test_0120", additional_paths=["dependencies"])
        datatypes_tuple = (
            self.url,
            datatypes_repository.name,
            datatypes_repository.owner,
            self.get_repository_tip(datatypes_repository),
        )
        self.create_repository_dependency(
            repository=tool_repository, repository_tuples=[datatypes_tuple], filepath=dependency_xml_path
        )

    def test_0040_verify_repository_dependency(self):
        """Verify the created repository dependency.

        We are at step 3a.
        Check the newly created repository dependency to ensure that it was defined and displays correctly.
        """
        datatypes_repository = self._get_repository_by_name_and_owner(
            datatypes_repository_name, common.test_user_2_name
        )
        tool_repository = self._get_repository_by_name_and_owner(tool_repository_name, common.test_user_1_name)
        self.check_repository_dependency(tool_repository, datatypes_repository)



import logging

import tool_shed.util.shed_util_common as suc
from galaxy import (
    util,
    web,
)
from tool_shed.managers.users import (
    api_create_user,
    index,
)
from tool_shed_client.schema import CreateUserRequest
from . import BaseShedAPIController

log = logging.getLogger(__name__)


class UsersController(BaseShedAPIController):
    """RESTful controller for interactions with users in the Tool Shed."""

    @web.expose_api
    @web.require_admin
    def create(self, trans, payload, **kwd):
        """
                POST /api/users
                Returns a dictionary of information about the created user.

        :       param key: the current Galaxy admin user's API key

                The following parameters are included in the payload.
                :param email (required): the email address of the user
                :param password (required): the password of the user
                :param username (required): the public username of the user
        """
        # Get the information about the user to be created from the payload.
        email = payload.get("email", "")
        password = payload.get("password", "")
        username = payload.get("username", "")
        # Create the user.
        request = CreateUserRequest(
            email=email,
            username=username,
            password=password,
        )
        user = api_create_user(trans, request)
        user_dict = user.dict()
        user_dict["message"] = f"User '{str(user.username)}' has been created."
        user_dict["url"] = web.url_for(controller="users", action="show", id=trans.security.encode_id(user.id))
        return user_dict

    def __get_value_mapper(self, trans):
        value_mapper = {"id": trans.security.encode_id}
        return value_mapper

    @web.expose_api_anonymous_and_sessionless
    def index(self, trans, deleted=False, **kwd):
        """
        GET /api/users
        Returns a list of dictionaries that contain information about each user.
        """
        # Example URL: http://localhost:9009/api/users
        user_dicts = []
        deleted = util.asbool(deleted)
        for user in index(trans.app, deleted):
            user_dict = user.dict()
            user_dict["url"] = web.url_for(controller="users", action="show", id=trans.security.encode_id(user.id))
            user_dicts.append(user_dict)
        return user_dicts

    @web.expose_api_anonymous_and_sessionless
    def show(self, trans, id, **kwd):
        """
        GET /api/users/{encoded_user_id}
        GET /api/users/current
        Returns a dictionary of information about a user.

        :param id: the encoded id of the User object.
        """
        user = None
        # user is requesting data about themselves
        user = trans.user if id == "current" else suc.get_user(trans.app, id)
        if user is None:
            user_dict = dict(message=f"Unable to locate user record for id {str(id)}.", status="error")
            return user_dict
        user_dict = user.to_dict(view="element", value_mapper=self.__get_value_mapper(trans))
        user_dict["url"] = web.url_for(controller="users", action="show", id=trans.security.encode_id(user.id))
        return user_dict



import logging
from typing import (
    Callable,
    Dict,
)

from sqlalchemy import select

from galaxy import (
    util,
    web,
)
from galaxy.exceptions import (
    AdminRequiredException,
    ObjectNotFound,
    RequestParameterMissingException,
)
from galaxy.util import (
    pretty_print_time_interval,
    UNKNOWN,
)
from galaxy.web import (
    expose_api,
    expose_api_anonymous_and_sessionless,
    require_admin,
)
from tool_shed.managers import groups
from tool_shed.structured_app import ToolShedApp
from tool_shed.webapp.model import (
    Category,
    Repository,
    RepositoryCategoryAssociation,
    RepositoryMetadata,
    User,
)
from . import BaseShedAPIController

log = logging.getLogger(__name__)


class GroupsController(BaseShedAPIController):
    """RESTful controller for interactions with groups in the Tool Shed."""

    def __init__(self, app: ToolShedApp):
        super().__init__(app)
        self.group_manager = groups.GroupManager()

    def __get_value_mapper(self, trans) -> Dict[str, Callable]:
        value_mapper = {"id": trans.security.encode_id}
        return value_mapper

    @expose_api_anonymous_and_sessionless
    def index(self, trans, deleted=False, **kwd):
        """
        GET /api/groups
        Return a list of dictionaries that contain information about each Group.

        :param deleted: flag used to include deleted groups

        Example: GET localhost:9009/api/groups
        """
        group_dicts = []
        deleted = util.asbool(deleted)
        if deleted and not trans.user_is_admin:
            raise AdminRequiredException("Only administrators can query deleted groups.")
        for group in self.group_manager.list(trans, deleted):
            group_dicts.append(self._populate(trans, group))
        return group_dicts

    @expose_api
    @require_admin
    def create(self, trans, payload, **kwd):
        """
        POST /api/groups
        Return a dictionary of information about the created group.
        The following parameters are included in the payload:

        :param name (required): the name of the group
        :param description (optional): the description of the group

        Example: POST /api/groups/?key=XXXYYYXXXYYY
        Content-Disposition: form-data; name="name" Group_Name
        Content-Disposition: form-data; name="description" Group_Description
        """
        group_dict = dict(message="", status="ok")
        if name := payload.get("name", ""):
            description = payload.get("description", "")
            if not description:
                description = ""
            else:
                # TODO add description field to the model
                group_dict = self.group_manager.create(trans, name=name).to_dict(
                    view="element", value_mapper=self.__get_value_mapper(trans)
                )
        else:
            raise RequestParameterMissingException('Missing required parameter "name".')
        return group_dict

    @expose_api_anonymous_and_sessionless
    def show(self, trans, encoded_id, **kwd):
        """
        GET /api/groups/{encoded_group_id}
        Return a dictionary of information about a group.

        :param id: the encoded id of the Group object

        Example: GET localhost:9009/api/groups/f9cad7b01a472135
        """
        decoded_id = trans.security.decode_id(encoded_id)
        group = self.group_manager.get(trans, decoded_id)
        if group is None:
            raise ObjectNotFound("Unable to locate group record with the given id.")
        return self._populate(trans, group)

    def _populate(self, trans, group):
        """
        Turn the given group information from DB into a dict
        and add other characteristics like members and repositories.
        """
        group_dict = group.to_dict(view="collection", value_mapper=self.__get_value_mapper(trans))
        group_members = []
        group_repos = []
        total_downloads = 0
        for uga in group.users:
            user = trans.sa_session.get(User, uga.user_id)
            user_repos_count = 0
            for repo in get_user_repositories(trans.sa_session, uga.user_id):
                categories = []
                for rca in repo.categories:
                    cat_dict = dict(name=rca.category.name, id=trans.app.security.encode_id(rca.category.id))
                    categories.append(cat_dict)
                time_repo_created_full = repo.create_time.strftime("%Y-%m-%d %I:%M %p")
                time_repo_updated_full = repo.update_time.strftime("%Y-%m-%d %I:%M %p")
                time_repo_created = pretty_print_time_interval(repo.create_time, True)
                time_repo_updated = pretty_print_time_interval(repo.update_time, True)
                # TODO add user ratings
                total_downloads += repo.times_downloaded
                group_repos.append(
                    {
                        "name": repo.name,
                        "times_downloaded": repo.times_downloaded,
                        "owner": repo.user.username,
                        "time_created_full": time_repo_created_full,
                        "time_created": time_repo_created,
                        "time_updated_full": time_repo_updated_full,
                        "time_updated": time_repo_updated,
                        "description": repo.description,
                        "categories": categories,
                    }
                )
                user_repos_count += 1
            encoded_user_id = trans.app.security.encode_id(user.id)
            user_repos_url = web.url_for(
                controller="repository", action="browse_repositories_by_user", user_id=encoded_user_id
            )
            time_created = pretty_print_time_interval(user.create_time, True)
            member_dict = {
                "id": encoded_user_id,
                "username": user.username,
                "user_repos_url": user_repos_url,
                "user_repos_count": user_repos_count,
                "user_tools_count": UNKNOWN,
                "time_created": time_created,
            }
            group_members.append(member_dict)
        group_dict["members"] = group_members
        group_dict["total_members"] = len(group_members)
        group_dict["repositories"] = group_repos
        group_dict["total_repos"] = len(group_repos)
        group_dict["total_downloads"] = total_downloads
        return group_dict


def get_user_repositories(session, user_id):
    stmt = (
        select(Repository)
        .where(Repository.user_id == user_id)
        .join(RepositoryMetadata)
        .join(User)
        .outerjoin(RepositoryCategoryAssociation)
        .outerjoin(Category)
    )
    return session.scalars(stmt)
